from __future__ import annotations

import re
from collections.abc import Callable, Iterator
from re import Match, Pattern


MatchEvaluator = Callable[[Match[str]], str]


class GroupCollection:
    def __init__(self, groups: list[str], named_groups: dict[str, str]) -> None:
        self.named_groups = named_groups
        self.groups = groups

    def values(self) -> list[str]:
        return list(self.groups)

    def __len__(self) -> int:
        return len(self.groups)

    def __getitem__(self, key: int | str) -> str | None:
        if isinstance(key, int):
            return self.groups[key]
        else:
            return self.named_groups.get(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.groups)


def _options_to_flags(options: int) -> int:
    """Convert .NET options to Python flags."""
    flags = 256
    if options & 1:
        flags |= re.IGNORECASE
    if options & 2:
        flags |= re.MULTILINE
    if options & 16:
        flags |= re.S
    return flags


def create(pattern: str, options: int = 0) -> Pattern[str]:
    # raw_pattern = pattern.encode("unicode_escape").decode("utf-8")
    flags = _options_to_flags(options)
    pattern = pattern.replace("?<", "?P<")
    return re.compile(pattern, flags=flags)


def escape(string: str) -> str:
    return re.escape(string)


def unescape(string: str) -> str:
    return re.sub(r"\\(.)", r"\1", string)


def match(reg: Pattern[str], input: str, start_at: int = 0) -> Match[str] | None:
    return reg.search(input, pos=start_at)


def matches(reg: Pattern[str], input: str, start_at: int = 0) -> list[Match[str]]:
    return list(reg.finditer(input, pos=start_at))


def is_match(reg: Pattern[str], input: str, start_at: int = 0) -> bool:
    return reg.search(input, pos=start_at) is not None


def groups(m: Match[str]) -> GroupCollection:
    named_groups: dict[str, str] = m.groupdict()

    # .NET adds the whole capture as group 0
    groups_ = [m.group(0), *m.groups()]

    return GroupCollection(named_groups=named_groups, groups=groups_)


def options(reg: Pattern[str]) -> int:
    options = 256  # ECMAScript
    if reg.flags & re.IGNORECASE:
        options |= 1
    if reg.flags & re.MULTILINE:
        options |= 2
    if reg.flags & re.S:
        options |= 16
    return options


def replace(
    reg: str | Pattern[str],
    input: str,
    replacement: str | MatchEvaluator,
    limit: int | None = None,
    offset: int = 0,
) -> str:
    if isinstance(replacement, str):
        # Rewrite placeholders from .NET e.g $1 to \g<1> Python syntax
        replacement = re.sub(pattern=r"\$(\d+)", repl=r"\\g<\g<1>>", string=replacement)

    if isinstance(reg, str):
        return re.sub(input, replacement, reg, count=limit or 0)
    else:
        return input[:offset] + reg.sub(replacement, input[offset:], count=limit or 0)


def split(
    reg: str | Pattern[str],
    input: str,
    limit: int | None = None,
    offset: int = 0,
) -> list[str]:
    if isinstance(reg, str):
        return re.split(input, reg, maxsplit=limit or 0)

    input = input[offset:]
    return reg.split(input, maxsplit=limit or 0)[:limit]


def get_item(groups: GroupCollection, index: str | int) -> str | None:
    return groups[index]


__all__ = [
    "escape",
    "is_match",
    "match",
    "matches",
    "create",
    "options",
    "replace",
    "split",
    "unescape",
    "groups",
    "get_item",
]



from __future__ import annotations
from ...arctrl_core.comment import Remark
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...fable_library.list import (of_array, FSharpList)
from ...fable_library.util import (IEnumerable_1, IEnumerator)
from .ontology_annotation import (from_sparse_table as from_sparse_table_1, to_sparse_table as to_sparse_table_1, from_rows as from_rows_1, to_rows as to_rows_1)
from .sparse_table import SparseTable

design_type_label: str = "Type"

design_type_term_accession_number_label: str = "Type Term Accession Number"

design_type_term_source_reflabel: str = "Type Term Source REF"

labels: FSharpList[str] = of_array([design_type_label, design_type_term_accession_number_label, design_type_term_source_reflabel])

def from_sparse_table(matrix: SparseTable) -> FSharpList[OntologyAnnotation]:
    return from_sparse_table_1(design_type_label, design_type_term_source_reflabel, design_type_term_accession_number_label, matrix)


def to_sparse_table(designs: FSharpList[OntologyAnnotation]) -> SparseTable:
    return to_sparse_table_1(design_type_label, design_type_term_source_reflabel, design_type_term_accession_number_label, designs)


def from_rows(prefix: str | None, line_number: int, rows: IEnumerator[IEnumerable_1[tuple[int, str]]]) -> tuple[str | None, int, FSharpList[Remark], FSharpList[OntologyAnnotation]]:
    return from_rows_1(prefix, design_type_label, design_type_term_source_reflabel, design_type_term_accession_number_label, line_number, rows)


def to_rows(prefix: str | None, designs: FSharpList[OntologyAnnotation]) -> IEnumerable_1[IEnumerable_1[tuple[int, str]]]:
    return to_rows_1(prefix, design_type_label, design_type_term_source_reflabel, design_type_term_accession_number_label, designs)


__all__ = ["design_type_label", "design_type_term_accession_number_label", "design_type_term_source_reflabel", "labels", "from_sparse_table", "to_sparse_table", "from_rows", "to_rows"]




from __future__ import annotations
from dataclasses import dataclass
from typing import (Any, TypeVar)
from ....fable_library.reflection import (TypeInfo, string_type, record_type)
from ....fable_library.seq import map
from ....fable_library.types import Record
from ....fable_library.util import (to_enumerable, IEnumerable_1)
from ....thoth_json_core.types import (IEncodable, IEncoderHelpers_1)

__A_ = TypeVar("__A_")

def _expr1804() -> TypeInfo:
    return record_type("ARCtrl.Json.ROCrateContext.Publication.IContext", [], IContext, lambda: [("sdo", string_type), ("arc", string_type), ("Publication", string_type), ("pub_med_id", string_type), ("doi", string_type), ("title", string_type), ("status", string_type), ("author_list", string_type), ("comments", string_type)])


@dataclass(eq = False, repr = False, slots = True)
class IContext(Record):
    sdo: str
    arc: str
    Publication: str
    pub_med_id: str
    doi: str
    title: str
    status: str
    author_list: str
    comments: str

IContext_reflection = _expr1804

def _arrow1814(__unit: None=None) -> IEncodable:
    class ObjectExpr1805(IEncodable):
        def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
            return helpers.encode_string("http://schema.org/")

    class ObjectExpr1806(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
            return helpers_1.encode_string("sdo:ScholarlyArticle")

    class ObjectExpr1807(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
            return helpers_2.encode_string("sdo:url")

    class ObjectExpr1808(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
            return helpers_3.encode_string("sdo:sameAs")

    class ObjectExpr1809(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any]) -> Any:
            return helpers_4.encode_string("sdo:headline")

    class ObjectExpr1810(IEncodable):
        def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
            return helpers_5.encode_string("sdo:creativeWorkStatus")

    class ObjectExpr1811(IEncodable):
        def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
            return helpers_6.encode_string("sdo:author")

    class ObjectExpr1812(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any]) -> Any:
            return helpers_7.encode_string("sdo:disambiguatingDescription")

    values: IEnumerable_1[tuple[str, IEncodable]] = to_enumerable([("sdo", ObjectExpr1805()), ("Publication", ObjectExpr1806()), ("pubMedID", ObjectExpr1807()), ("doi", ObjectExpr1808()), ("title", ObjectExpr1809()), ("status", ObjectExpr1810()), ("authorList", ObjectExpr1811()), ("comments", ObjectExpr1812())])
    class ObjectExpr1813(IEncodable):
        def Encode(self, helpers_8: IEncoderHelpers_1[Any]) -> Any:
            def mapping(tupled_arg: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg[0], tupled_arg[1].Encode(helpers_8))

            arg: IEnumerable_1[tuple[str, __A_]] = map(mapping, values)
            return helpers_8.encode_object(arg)

    return ObjectExpr1813()


context_jsonvalue: IEncodable = _arrow1814()

__all__ = ["IContext_reflection", "context_jsonvalue"]




from __future__ import annotations
from typing import Any
from ..fable_library.option import map
from ..fable_library.reflection import (TypeInfo, class_type)
from ..fable_library.types import Array
from .comment import Comment
from .data import (Data, Data_reflection)
from .data_file import DataFile
from .Helper.hash_codes import (box_hash_array, box_hash_option, box_hash_seq, hash_1)
from .ontology_annotation import OntologyAnnotation
from .value import Value as Value_1

def _expr729() -> TypeInfo:
    return class_type("ARCtrl.DataContext", None, DataContext, Data_reflection())


class DataContext(Data):
    def __init__(self, id: str | None=None, name: str | None=None, data_type: DataFile | None=None, format: str | None=None, selector_format: str | None=None, explication: OntologyAnnotation | None=None, unit: OntologyAnnotation | None=None, object_type: OntologyAnnotation | None=None, label: str | None=None, description: str | None=None, generated_by: str | None=None, comments: Array[Comment] | None=None) -> None:
        super().__init__(id, name, data_type, format, selector_format, comments)
        self._explication: OntologyAnnotation | None = explication
        self._unit: OntologyAnnotation | None = unit
        self._objectType: OntologyAnnotation | None = object_type
        self._label: str | None = label
        self._description: str | None = description
        self._generatedBy: str | None = generated_by

    def __hash__(self, __unit: None=None) -> Any:
        this: DataContext = self
        return box_hash_array([box_hash_option(this.ID), box_hash_option(this.Name), box_hash_option(this.DataType), box_hash_option(this.Format), box_hash_option(this.SelectorFormat), box_hash_seq(this.Comments), box_hash_option(DataContext__get_Explication(this)), box_hash_option(DataContext__get_Unit(this)), box_hash_option(DataContext__get_ObjectType(this)), box_hash_option(DataContext__get_Label(this)), box_hash_option(DataContext__get_Description(this)), box_hash_option(DataContext__get_GeneratedBy(this))])

    def __eq__(self, obj: Any=None) -> bool:
        this: DataContext = self
        return hash_1(this) == hash_1(obj)

    def AlternateName(self, __unit: None=None) -> str | None:
        this: DataContext = self
        return DataContext__get_Label(this)

    def MeasurementMethod(self, __unit: None=None) -> str | None:
        this: DataContext = self
        return DataContext__get_GeneratedBy(this)

    def GetCategory(self, __unit: None=None) -> OntologyAnnotation | None:
        this: DataContext = self
        return DataContext__get_Explication(this)

    def GetValue(self, __unit: None=None) -> Value_1 | None:
        this: DataContext = self
        def mapping(Item: OntologyAnnotation) -> Value_1:
            return Value_1(0, Item)

        return map(mapping, DataContext__get_ObjectType(this))

    def GetUnit(self, __unit: None=None) -> OntologyAnnotation | None:
        this: DataContext = self
        return DataContext__get_Unit(this)

    def GetAdditionalType(self, __unit: None=None) -> str:
        return "DataContext"

    def Description(self, __unit: None=None) -> str | None:
        this: DataContext = self
        return DataContext__get_Description(this)


DataContext_reflection = _expr729

def DataContext__ctor_Z780A8A2A(id: str | None=None, name: str | None=None, data_type: DataFile | None=None, format: str | None=None, selector_format: str | None=None, explication: OntologyAnnotation | None=None, unit: OntologyAnnotation | None=None, object_type: OntologyAnnotation | None=None, label: str | None=None, description: str | None=None, generated_by: str | None=None, comments: Array[Comment] | None=None) -> DataContext:
    return DataContext(id, name, data_type, format, selector_format, explication, unit, object_type, label, description, generated_by, comments)


def DataContext__get_Explication(this: DataContext) -> OntologyAnnotation | None:
    return this._explication


def DataContext__set_Explication_279AAFF2(this: DataContext, explication: OntologyAnnotation | None=None) -> None:
    this._explication = explication


def DataContext__get_Unit(this: DataContext) -> OntologyAnnotation | None:
    return this._unit


def DataContext__set_Unit_279AAFF2(this: DataContext, unit: OntologyAnnotation | None=None) -> None:
    this._unit = unit


def DataContext__get_ObjectType(this: DataContext) -> OntologyAnnotation | None:
    return this._objectType


def DataContext__set_ObjectType_279AAFF2(this: DataContext, object_type: OntologyAnnotation | None=None) -> None:
    this._objectType = object_type


def DataContext__get_Label(this: DataContext) -> str | None:
    return this._label


def DataContext__set_Label_6DFDD678(this: DataContext, label: str | None=None) -> None:
    this._label = label


def DataContext__get_Description(this: DataContext) -> str | None:
    return this._description


def DataContext__set_Description_6DFDD678(this: DataContext, description: str | None=None) -> None:
    this._description = description


def DataContext__get_GeneratedBy(this: DataContext) -> str | None:
    return this._generatedBy


def DataContext__set_GeneratedBy_6DFDD678(this: DataContext, generated_by: str | None=None) -> None:
    this._generatedBy = generated_by


def DataContext__AsData(this: DataContext) -> Data:
    return Data(this.ID, this.Name, this.DataType, this.Format, this.SelectorFormat, this.Comments)


def DataContext_fromData_Z7B4D7BF5(data: Data, explication: OntologyAnnotation | None=None, unit: OntologyAnnotation | None=None, object_type: OntologyAnnotation | None=None, label: str | None=None, description: str | None=None, generated_by: str | None=None) -> DataContext:
    return DataContext__ctor_Z780A8A2A(data.ID, data.Name, data.DataType, data.Format, data.SelectorFormat, explication, unit, object_type, label, description, generated_by, data.Comments)


def DataContext_createAsPV(alternate_name: str | None=None, measurement_method: str | None=None, description: str | None=None, category: OntologyAnnotation | None=None, value: Value_1 | None=None, unit: OntologyAnnotation | None=None) -> DataContext:
    def _arrow730(__unit: None=None, alternate_name: Any=alternate_name, measurement_method: Any=measurement_method, description: Any=description, category: Any=category, value: Any=value, unit: Any=unit) -> OntologyAnnotation | None:
        oa: OntologyAnnotation = value.fields[0]
        return oa

    def _arrow731(__unit: None=None, alternate_name: Any=alternate_name, measurement_method: Any=measurement_method, description: Any=description, category: Any=category, value: Any=value, unit: Any=unit) -> OntologyAnnotation | None:
        v: Value_1 = value
        return OntologyAnnotation(v.Text)

    return DataContext__ctor_Z780A8A2A(None, None, None, None, None, category, unit, None if (value is None) else (_arrow730() if (value.tag == 0) else _arrow731()), alternate_name, description, measurement_method)


def DataContext__Copy(this: DataContext) -> DataContext:
    copy: DataContext = DataContext__ctor_Z780A8A2A()
    copy.ID = this.ID
    copy.Name = this.Name
    copy.DataType = this.DataType
    copy.Format = this.Format
    copy.SelectorFormat = this.SelectorFormat
    DataContext__set_Explication_279AAFF2(copy, DataContext__get_Explication(this))
    DataContext__set_Unit_279AAFF2(copy, DataContext__get_Unit(this))
    DataContext__set_ObjectType_279AAFF2(copy, DataContext__get_ObjectType(this))
    DataContext__set_Description_6DFDD678(copy, DataContext__get_Description(this))
    DataContext__set_GeneratedBy_6DFDD678(copy, DataContext__get_GeneratedBy(this))
    copy.Comments = this.Comments
    return copy


__all__ = ["DataContext_reflection", "DataContext__get_Explication", "DataContext__set_Explication_279AAFF2", "DataContext__get_Unit", "DataContext__set_Unit_279AAFF2", "DataContext__get_ObjectType", "DataContext__set_ObjectType_279AAFF2", "DataContext__get_Label", "DataContext__set_Label_6DFDD678", "DataContext__get_Description", "DataContext__set_Description_6DFDD678", "DataContext__get_GeneratedBy", "DataContext__set_GeneratedBy_6DFDD678", "DataContext__AsData", "DataContext_fromData_Z7B4D7BF5", "DataContext_createAsPV", "DataContext__Copy"]




from __future__ import annotations
from typing import Any
from ..fable_library.date import now
from ..fable_library.guid import new_guid
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, string_type, union_type, class_type)
from ..fable_library.types import (Array, Union)
from ..fable_library.util import (equals, safe_hash)
from .Helper.collections_ import ResizeArray_map
from .Helper.hash_codes import (box_hash_array, box_hash_seq, hash_date_time)
from .Helper.sem_ver import (SemVer_tryOfString_Z721C83C5, SemVer)
from .ontology_annotation import OntologyAnnotation
from .person import Person
from .Table.arc_table import ArcTable

def _expr974() -> TypeInfo:
    return union_type("ARCtrl.Organisation", [], Organisation, lambda: [[], [("Item", string_type)]])


class Organisation(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["DataPLANT", "Other"]

    @staticmethod
    def of_string(str_1: str) -> Organisation:
        return Organisation(0) if (str_1.lower() == "dataplant") else Organisation(1, str_1)

    def __str__(self, __unit: None=None) -> str:
        this: Organisation = self
        return this.fields[0] if (this.tag == 1) else "DataPLANT"

    def IsOfficial(self, __unit: None=None) -> bool:
        this: Organisation = self
        return equals(this, Organisation(0))


Organisation_reflection = _expr974

def _expr976() -> TypeInfo:
    return class_type("ARCtrl.Template", None, Template)


class Template:
    def __init__(self, id: str, table: ArcTable, name: str | None=None, description: str | None=None, organisation: Organisation | None=None, version: str | None=None, authors: Array[Person] | None=None, repos: Array[OntologyAnnotation] | None=None, tags: Array[OntologyAnnotation] | None=None, last_updated: Any | None=None) -> None:
        name_1: str = default_arg(name, "")
        description_1: str = default_arg(description, "")
        organisation_1: Organisation = default_arg(organisation, Organisation(1, "Custom Organisation"))
        version_1: str = default_arg(version, "0.0.0")
        authors_1: Array[Person] = default_arg(authors, [])
        repos_1: Array[OntologyAnnotation] = default_arg(repos, [])
        tags_1: Array[OntologyAnnotation] = default_arg(tags, [])
        last_updated_1: Any = default_arg(last_updated, now())
        self._Id: str = id
        self._Table: ArcTable = table
        self._Name: str = name_1
        self._Description: str = description_1
        self._Organisation: Organisation = organisation_1
        self._Version: str = version_1
        self._Authors: Array[Person] = authors_1
        self._EndpointRepositories: Array[OntologyAnnotation] = repos_1
        self._Tags: Array[OntologyAnnotation] = tags_1
        self._LastUpdated: Any = last_updated_1

    @property
    def Id(self, __unit: None=None) -> str:
        __: Template = self
        return __._Id

    @Id.setter
    def Id(self, v: str) -> None:
        __: Template = self
        __._Id = v

    @property
    def Table(self, __unit: None=None) -> ArcTable:
        __: Template = self
        return __._Table

    @Table.setter
    def Table(self, v: ArcTable) -> None:
        __: Template = self
        __._Table = v

    @property
    def Name(self, __unit: None=None) -> str:
        __: Template = self
        return __._Name

    @Name.setter
    def Name(self, v: str) -> None:
        __: Template = self
        __._Name = v

    @property
    def Description(self, __unit: None=None) -> str:
        __: Template = self
        return __._Description

    @Description.setter
    def Description(self, v: str) -> None:
        __: Template = self
        __._Description = v

    @property
    def Organisation(self, __unit: None=None) -> Organisation:
        __: Template = self
        return __._Organisation

    @Organisation.setter
    def Organisation(self, v: Organisation) -> None:
        __: Template = self
        __._Organisation = v

    @property
    def Version(self, __unit: None=None) -> str:
        __: Template = self
        return __._Version

    @Version.setter
    def Version(self, v: str) -> None:
        __: Template = self
        __._Version = v

    @property
    def Authors(self, __unit: None=None) -> Array[Person]:
        __: Template = self
        return __._Authors

    @Authors.setter
    def Authors(self, v: Array[Person]) -> None:
        __: Template = self
        __._Authors = v

    @property
    def EndpointRepositories(self, __unit: None=None) -> Array[OntologyAnnotation]:
        __: Template = self
        return __._EndpointRepositories

    @EndpointRepositories.setter
    def EndpointRepositories(self, v: Array[OntologyAnnotation]) -> None:
        __: Template = self
        __._EndpointRepositories = v

    @property
    def Tags(self, __unit: None=None) -> Array[OntologyAnnotation]:
        __: Template = self
        return __._Tags

    @Tags.setter
    def Tags(self, v: Array[OntologyAnnotation]) -> None:
        __: Template = self
        __._Tags = v

    @property
    def LastUpdated(self, __unit: None=None) -> Any:
        __: Template = self
        return __._LastUpdated

    @LastUpdated.setter
    def LastUpdated(self, v: Any) -> None:
        __: Template = self
        __._LastUpdated = v

    @staticmethod
    def make(id: str, table: ArcTable, name: str, description: str, organisation: Organisation, version: str, authors: Array[Person], repos: Array[OntologyAnnotation], tags: Array[OntologyAnnotation], last_updated: Any) -> Template:
        return Template(id, table, name, description, organisation, version, authors, repos, tags, last_updated)

    @staticmethod
    def create(id: str, table: ArcTable, name: str | None=None, description: str | None=None, organisation: Organisation | None=None, version: str | None=None, authors: Array[Person] | None=None, repos: Array[OntologyAnnotation] | None=None, tags: Array[OntologyAnnotation] | None=None, last_updated: Any | None=None) -> Template:
        return Template(id, table, name, description, organisation, version, authors, repos, tags, last_updated)

    @staticmethod
    def init(template_name: str) -> Template:
        return Template(new_guid(), ArcTable.init(template_name), template_name)

    @property
    def SemVer(self, __unit: None=None) -> SemVer | None:
        this: Template = self
        return SemVer_tryOfString_Z721C83C5(this.Version)

    def ReferenceEquals(self, other: Template) -> bool:
        this: Template = self
        return this is other

    def StructurallyEquals(self, other: Template) -> bool:
        this: Template = self
        return safe_hash(this) == safe_hash(other)

    def Copy(self, __unit: None=None) -> Template:
        this: Template = self
        def f(c: Person) -> Person:
            return c.Copy()

        next_authors: Array[Person] = ResizeArray_map(f, this.Authors)
        def f_1(c_1: OntologyAnnotation) -> OntologyAnnotation:
            return c_1.Copy()

        next_repos: Array[OntologyAnnotation] = ResizeArray_map(f_1, this.EndpointRepositories)
        def f_2(c_2: OntologyAnnotation) -> OntologyAnnotation:
            return c_2.Copy()

        next_tags: Array[OntologyAnnotation] = ResizeArray_map(f_2, this.Tags)
        return Template.create(this.Id, this.Table.Copy(), this.Name, this.Description, this.Organisation, this.Version, next_authors, next_repos, next_tags, this.LastUpdated)

    def __eq__(self, other: Any=None) -> bool:
        this: Template = self
        return this.StructurallyEquals(other) if isinstance(other, Template) else False

    def __hash__(self, __unit: None=None) -> Any:
        this: Template = self
        def _arrow975(__unit: None=None) -> str:
            copy_of_struct: str = this.Id
            return str(copy_of_struct)

        return box_hash_array([_arrow975(), safe_hash(this.Table), this.Name, safe_hash(this.Organisation), this.Version, box_hash_seq(this.Authors), box_hash_seq(this.EndpointRepositories), box_hash_seq(this.Tags), hash_date_time(this.LastUpdated)])


Template_reflection = _expr976

def Template__ctor_Z17EA765C(id: str, table: ArcTable, name: str | None=None, description: str | None=None, organisation: Organisation | None=None, version: str | None=None, authors: Array[Person] | None=None, repos: Array[OntologyAnnotation] | None=None, tags: Array[OntologyAnnotation] | None=None, last_updated: Any | None=None) -> Template:
    return Template(id, table, name, description, organisation, version, authors, repos, tags, last_updated)


__all__ = ["Organisation_reflection", "Template_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import (value as value_1, default_arg)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import (FSharpRef, Array)
from ..ldobject import (LDNode, LDNode_reflection)

def _expr1576() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.PropertyValue", None, PropertyValue, LDNode_reflection())


class PropertyValue(LDNode):
    def __init__(self, id: str, name: Any=None, value: Any=None, property_id: Any | None=None, unit_code: Any | None=None, unit_text: Any | None=None, value_reference: Any | None=None, additional_type: Array[str] | None=None) -> None:
        super().__init__(id, ["schema.org/PropertyValue"], default_arg(additional_type, []))
        this: FSharpRef[PropertyValue] = FSharpRef(None)
        this.contents = self
        self.init_00408: int = 1
        set_property("name", name, this.contents)
        set_property("value", value, this.contents)
        set_optional_property("propertyID", property_id, this.contents)
        set_optional_property("unitCode", unit_code, this.contents)
        set_optional_property("unitText", unit_text, this.contents)
        set_optional_property("valueReference", value_reference, this.contents)

    def GetName(self, __unit: None=None) -> str:
        this: PropertyValue = self
        obj: DynamicObj = this
        if try_get_property_value("name", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("name")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "name") + "\' is set on this \'") + "PropertyValue") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "name") + "\' set on this \'") + "PropertyValue") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_name() -> Callable[[PropertyValue], str]:
        def _arrow1574(lp: PropertyValue) -> str:
            return lp.GetName()

        return _arrow1574

    def GetValue(self, __unit: None=None) -> str:
        this: PropertyValue = self
        obj: DynamicObj = this
        if try_get_property_value("name", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("name")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "name") + "\' is set on this \'") + "PropertyValue") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "name") + "\' set on this \'") + "PropertyValue") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_value() -> Callable[[PropertyValue], str]:
        def _arrow1575(lp: PropertyValue) -> str:
            return lp.GetValue()

        return _arrow1575


PropertyValue_reflection = _expr1576

def PropertyValue__ctor_Z5E5247A6(id: str, name: Any=None, value: Any=None, property_id: Any | None=None, unit_code: Any | None=None, unit_text: Any | None=None, value_reference: Any | None=None, additional_type: Array[str] | None=None) -> PropertyValue:
    return PropertyValue(id, name, value, property_id, unit_code, unit_text, value_reference, additional_type)


__all__ = ["PropertyValue_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ....arctrl_core.Process.process_input import ProcessInput
from ....arctrl_json.encode import default_spaces
from ....arctrl_json.Process.process_input import (ISAJson_decoder, ISAJson_encoder)
from ....fable_library.option import default_arg
from ....fable_library.result import FSharpResult_2
from ....fable_library.string_ import (to_text, printf)
from ....thoth_json_core.types import IEncodable
from ....thoth_json_python.decode import Decode_fromString
from ....thoth_json_python.encode import to_string

def ARCtrl_Process_ProcessInput__ProcessInput_fromISAJsonString_Static_Z721C83C5(s: str) -> ProcessInput:
    match_value: FSharpResult_2[ProcessInput, str] = Decode_fromString(ISAJson_decoder, s)
    if match_value.tag == 1:
        raise Exception(to_text(printf("Error decoding string: %O"))(match_value.fields[0]))

    else: 
        return match_value.fields[0]



def ARCtrl_Process_ProcessInput__ProcessInput_toISAJsonString_Static_Z3B036AA(spaces: int | None=None, use_idreferencing: bool | None=None) -> Callable[[ProcessInput], str]:
    id_map: Any | None = dict([]) if default_arg(use_idreferencing, False) else None
    def _arrow3823(f: ProcessInput, spaces: Any=spaces, use_idreferencing: Any=use_idreferencing) -> str:
        value: IEncodable = ISAJson_encoder(id_map, f)
        return to_string(default_spaces(spaces), value)

    return _arrow3823


def ARCtrl_Process_ProcessInput__ProcessInput_ToISAJsonString_Z3B036AA(this: ProcessInput, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
    return ARCtrl_Process_ProcessInput__ProcessInput_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(this)


__all__ = ["ARCtrl_Process_ProcessInput__ProcessInput_fromISAJsonString_Static_Z721C83C5", "ARCtrl_Process_ProcessInput__ProcessInput_toISAJsonString_Static_Z3B036AA", "ARCtrl_Process_ProcessInput__ProcessInput_ToISAJsonString_Z3B036AA"]




from __future__ import annotations
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any
from ..fable_library.option import default_arg
from ..fable_library.reflection import (TypeInfo, int32_type, record_type, string_type, class_type, list_type, union_type, option_type)
from ..fable_library.string_ import initialize
from ..fable_library.system_text import (StringBuilder__ctor, StringBuilder__AppendLine_Z721C83C5)
from ..fable_library.types import (Record, to_string, Array, Union)
from ..fable_library.util import (ignore, get_enumerator)

def _expr392() -> TypeInfo:
    return record_type("YAMLicious.YAMLiciousTypes.Config", [], Config, lambda: [("Whitespace", int32_type), ("Level", int32_type)])


@dataclass(eq = False, repr = False, slots = True)
class Config(Record):
    Whitespace: int
    Level: int

Config_reflection = _expr392

def Config_init_71136F3F(whitespace: int | None=None) -> Config:
    return Config(default_arg(whitespace, 4), 0)


def Config__get_WhitespaceString(this: Config) -> str:
    def _arrow393(_arg: int, this: Any=this) -> str:
        return " "

    return initialize(this.Level * this.Whitespace, _arrow393)


def _expr394() -> TypeInfo:
    return record_type("YAMLicious.YAMLiciousTypes.Preprocessor", [], Preprocessor, lambda: [("AST", PreprocessorElement_reflection()), ("StringMap", class_type("System.Collections.Generic.Dictionary`2", [int32_type, string_type])), ("CommentMap", class_type("System.Collections.Generic.Dictionary`2", [int32_type, string_type]))])


@dataclass(eq = False, repr = False, slots = True)
class Preprocessor(Record):
    AST: PreprocessorElement
    StringMap: Any
    CommentMap: Any

Preprocessor_reflection = _expr394

def _expr396() -> TypeInfo:
    return union_type("YAMLicious.YAMLiciousTypes.PreprocessorElement", [], PreprocessorElement, lambda: [[("Item", list_type(PreprocessorElement_reflection()))], [("Item", list_type(PreprocessorElement_reflection()))], [("Item", string_type)], []])


class PreprocessorElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Level", "Intendation", "Line", "Nil"]

    def __str__(self, __unit: None=None) -> str:
        this: PreprocessorElement = self
        sb: Any = StringBuilder__ctor()
        def innerprint(next_1: PreprocessorElement, level: int) -> None:
            def _arrow395(_arg: int, next_1: Any=next_1, level: Any=level) -> str:
                return " "

            indent: str = initialize(level * 2, _arrow395)
            if next_1.tag == 1:
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, indent + "Intendation ["))
                with get_enumerator(next_1.fields[0]) as enumerator:
                    while enumerator.System_Collections_IEnumerator_MoveNext():
                        innerprint(enumerator.System_Collections_Generic_IEnumerator_1_get_Current(), level + 1)
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, indent + "]"))

            elif next_1.tag == 0:
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, indent + "Level ["))
                with get_enumerator(next_1.fields[0]) as enumerator_1:
                    while enumerator_1.System_Collections_IEnumerator_MoveNext():
                        innerprint(enumerator_1.System_Collections_Generic_IEnumerator_1_get_Current(), level + 1)
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, indent + "]"))

            elif next_1.tag == 3:
                pass

            else: 
                ignore(StringBuilder__AppendLine_Z721C83C5(sb, indent + (("Line \"" + next_1.fields[0]) + "\"")))


        innerprint(this, 0)
        return to_string(sb)


PreprocessorElement_reflection = _expr396

def _expr397() -> TypeInfo:
    return record_type("YAMLicious.YAMLiciousTypes.YAMLContent", [], YAMLContent, lambda: [("Value", string_type), ("Comment", option_type(string_type))])


@dataclass(eq = False, repr = False, slots = True)
class YAMLContent(Record):
    Value: str
    Comment: str | None

YAMLContent_reflection = _expr397

def YAMLContent_create_27AED5E3(value: str, comment: str | None=None) -> YAMLContent:
    return YAMLContent(value, comment)


def _expr398() -> TypeInfo:
    return union_type("YAMLicious.YAMLiciousTypes.YAMLElement", [], YAMLElement, lambda: [[("Item1", YAMLContent_reflection()), ("Item2", YAMLElement_reflection())], [("Item", YAMLContent_reflection())], [("Item", list_type(YAMLElement_reflection()))], [("Item", list_type(YAMLElement_reflection()))], [("Item", string_type)], []])


class YAMLElement(Union):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Mapping", "Value", "Sequence", "Object", "Comment", "Nil"]


YAMLElement_reflection = _expr398

__all__ = ["Config_reflection", "Config_init_71136F3F", "Config__get_WhitespaceString", "Preprocessor_reflection", "PreprocessorElement_reflection", "YAMLContent_reflection", "YAMLContent_create_27AED5E3", "YAMLElement_reflection"]




from __future__ import annotations
from collections.abc import Callable
from typing import Any
from ...dynamic_obj.dynamic_obj import DynamicObj
from ...dynamic_obj.dyn_obj import (try_get_property_value, set_property, set_optional_property)
from ...fable_library.option import value as value_1
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.types import FSharpRef
from .dataset import (Dataset, Dataset_reflection)

def _expr1547() -> TypeInfo:
    return class_type("ARCtrl.ROCrate.Assay", None, Assay, Dataset_reflection())


class Assay(Dataset):
    def __init__(self, id: str, identifier: str, about: Any | None=None, comment: Any | None=None, creator: Any | None=None, has_part: Any | None=None, measurement_method: Any | None=None, measurement_technique: Any | None=None, url: Any | None=None, variable_measured: Any | None=None) -> None:
        super().__init__(id, ["Assay"])
        this: FSharpRef[Assay] = FSharpRef(None)
        this.contents = self
        self.init_00408_1: int = 1
        set_property("identifier", identifier, this.contents)
        set_optional_property("measurementMethod", measurement_method, this.contents)
        set_optional_property("measurementTechnique", measurement_technique, this.contents)
        set_optional_property("variableMeasured", variable_measured, this.contents)
        set_optional_property("about", about, this.contents)
        set_optional_property("comment", comment, this.contents)
        set_optional_property("creator", creator, this.contents)
        set_optional_property("hasPart", has_part, this.contents)
        set_optional_property("url", url, this.contents)

    def GetIdentifier(self, __unit: None=None) -> str:
        this: Assay = self
        obj: DynamicObj = this
        if try_get_property_value("identifier", obj) is not None:
            match_value: str | None
            match_value_1: Any | None = obj.TryGetPropertyValue("identifier")
            if match_value_1 is not None:
                o: Any = value_1(match_value_1)
                match_value = o if (str(type(o)) == "<class \'str\'>") else None

            else: 
                match_value = None

            if match_value is None:
                raise Exception(((((("Property \'" + "identifier") + "\' is set on this \'") + "Assay") + "\' object but cannot be cast to \'") + "String") + "\'")

            else: 
                return match_value


        else: 
            raise Exception(((("No property \'" + "identifier") + "\' set on this \'") + "Assay") + "\' object although it is mandatory. Was it created correctly?")


    @staticmethod
    def get_identifier() -> Callable[[Assay], str]:
        def _arrow1546(ass: Assay) -> str:
            return ass.GetIdentifier()

        return _arrow1546


Assay_reflection = _expr1547

def Assay__ctor_Z318F9460(id: str, identifier: str, about: Any | None=None, comment: Any | None=None, creator: Any | None=None, has_part: Any | None=None, measurement_method: Any | None=None, measurement_technique: Any | None=None, url: Any | None=None, variable_measured: Any | None=None) -> Assay:
    return Assay(id, identifier, about, comment, creator, has_part, measurement_method, measurement_technique, url, variable_measured)


__all__ = ["Assay_reflection"]




from __future__ import annotations
from dataclasses import dataclass
from ...fable_library.option import default_arg
from ...fable_library.reflection import (TypeInfo, array_type, record_type)
from ...fable_library.seq import (exists, to_array, delay, collect, singleton, empty, is_empty, fold)
from ...fable_library.types import (Array, Record)
from ...fable_library.util import IEnumerable_1
from ..ontology_annotation import OntologyAnnotation
from .composite_cell import (CompositeCell, CompositeCell_reflection)
from .composite_header import (CompositeHeader, CompositeHeader_reflection)

def _expr817() -> TypeInfo:
    return record_type("ARCtrl.CompositeColumn", [], CompositeColumn, lambda: [("Header", CompositeHeader_reflection()), ("Cells", array_type(CompositeCell_reflection()))])


@dataclass(eq = False, repr = False, slots = True)
class CompositeColumn(Record):
    Header: CompositeHeader
    Cells: Array[CompositeCell]
    @staticmethod
    def create(header: CompositeHeader, cells: Array[CompositeCell] | None=None) -> CompositeColumn:
        return CompositeColumn(header, default_arg(cells, []))

    def Validate(self, raise_exception: bool | None=None) -> bool:
        this: CompositeColumn = self
        def predicate(c: CompositeCell) -> bool:
            return not c.ValidateAgainstHeader(this.Header, raise_exception)

        return not exists(predicate, this.Cells)

    def TryGetColumnUnits(self, __unit: None=None) -> Array[OntologyAnnotation] | None:
        this: CompositeColumn = self
        def _arrow816(__unit: None=None) -> IEnumerable_1[OntologyAnnotation]:
            def _arrow815(cell: CompositeCell) -> IEnumerable_1[OntologyAnnotation]:
                return singleton(cell.AsUnitized[1]) if cell.is_unitized else empty()

            return collect(_arrow815, this.Cells)

        arr: Array[OntologyAnnotation] = list(to_array(delay(_arrow816)))
        return None if is_empty(arr) else arr

    def GetDefaultEmptyCell(self, __unit: None=None) -> CompositeCell:
        this: CompositeColumn = self
        if not this.Header.IsTermColumn:
            return CompositeCell.empty_free_text()

        else: 
            pattern_input: tuple[int, int]
            arg: tuple[int, int] = (0, 0)
            def folder(tupled_arg: tuple[int, int], cell: CompositeCell) -> tuple[int, int]:
                units: int = tupled_arg[0] or 0
                terms: int = tupled_arg[1] or 0
                if cell.is_unitized:
                    return (units + 1, terms)

                else: 
                    return (units, terms + 1)


            pattern_input = fold(folder, (arg[0], arg[1]), this.Cells)
            return CompositeCell.empty_term() if (pattern_input[1] >= pattern_input[0]) else CompositeCell.empty_unitized()


    @property
    def IsUnique(self, __unit: None=None) -> bool:
        this: CompositeColumn = self
        return this.Header.IsUnique


CompositeColumn_reflection = _expr817

__all__ = ["CompositeColumn_reflection"]




from __future__ import annotations
from ..fable_library.list import of_array
from .yamlicious_types import PreprocessorElement

SequenceMappingsAST: PreprocessorElement = PreprocessorElement(0, of_array([PreprocessorElement(2, "-"), PreprocessorElement(1, of_array([PreprocessorElement(2, "My Key1: My Value1"), PreprocessorElement(2, "My Key2: My Value2"), PreprocessorElement(2, "My Key3: My Value3")])), PreprocessorElement(2, "-"), PreprocessorElement(1, of_array([PreprocessorElement(2, "My Key4: My Value4"), PreprocessorElement(2, "My Key5: My Value5"), PreprocessorElement(2, "My Key6: My Value6")]))]))

StringReplaceClean: str = "\r\nMy Key: </1>\r\n"

__all__ = ["SequenceMappingsAST", "StringReplaceClean"]




from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from .list import FSharpList, empty, singleton
from .option import some
from .reflection import TypeInfo, union_type
from .types import Array, Union
from .util import equals


_T = TypeVar("_T")

_TERROR = TypeVar("_TERROR")

_A = TypeVar("_A")

_B = TypeVar("_B")

_C = TypeVar("_C")

_S = TypeVar("_S")


def _expr34(gen0: TypeInfo, gen1: TypeInfo) -> TypeInfo:
    return union_type(
        "FSharp.Core.FSharpResult`2",
        [gen0, gen1],
        FSharpResult_2,
        lambda: [[("ResultValue", gen0)], [("ErrorValue", gen1)]],
    )


class FSharpResult_2(Union, Generic[_T, _TERROR]):
    def __init__(self, tag: int, *fields: Any) -> None:
        super().__init__()
        self.tag: int = tag or 0
        self.fields: Array[Any] = list(fields)

    @staticmethod
    def cases() -> list[str]:
        return ["Ok", "Error"]


FSharpResult_2_reflection = _expr34


def Result_Map(mapping: Callable[[_A], _B], result: FSharpResult_2[Any, Any]) -> FSharpResult_2[Any, Any]:
    if result.tag == 0:
        return FSharpResult_2(0, mapping(result.fields[0]))

    else:
        return FSharpResult_2(1, result.fields[0])


def Result_MapError(mapping: Callable[[_A], _B], result: FSharpResult_2[Any, Any]) -> FSharpResult_2[Any, Any]:
    if result.tag == 0:
        return FSharpResult_2(0, result.fields[0])

    else:
        return FSharpResult_2(1, mapping(result.fields[0]))


def Result_Bind(
    binder: Callable[[_A], FSharpResult_2[_B, _C]], result: FSharpResult_2[Any, Any]
) -> FSharpResult_2[Any, Any]:
    if result.tag == 0:
        return binder(result.fields[0])

    else:
        return FSharpResult_2(1, result.fields[0])


def Result_IsOk(result: FSharpResult_2[Any, Any]) -> bool:
    if result.tag == 0:
        return True

    else:
        return False


def Result_IsError(result: FSharpResult_2[Any, Any]) -> bool:
    if result.tag == 0:
        return False

    else:
        return True


def Result_Contains(value: Any, result: FSharpResult_2[Any, Any]) -> bool:
    if result.tag == 0:
        return equals(result.fields[0], value)

    else:
        return False


def Result_Count(result: FSharpResult_2[Any, Any]) -> int:
    if result.tag == 0:
        return 1

    else:
        return 0


def Result_DefaultValue(default_value: Any, result: FSharpResult_2[Any, Any]) -> Any:
    if result.tag == 0:
        return result.fields[0]

    else:
        return default_value


def Result_DefaultWith(def_thunk: Callable[[_B], _A], result: FSharpResult_2[Any, Any]) -> Any:
    if result.tag == 0:
        return result.fields[0]

    else:
        return def_thunk(result.fields[0])


def Result_Exists(predicate: Callable[[_A], bool], result: FSharpResult_2[Any, Any]) -> bool:
    if result.tag == 0:
        return predicate(result.fields[0])

    else:
        return False


def Result_Fold(folder: Callable[[_S, _A], _S], state: Any, result: FSharpResult_2[Any, Any]) -> Any:
    if result.tag == 0:
        return folder(state, result.fields[0])

    else:
        return state


def Result_FoldBack(folder: Callable[[_A, _S], _S], result: FSharpResult_2[Any, Any], state: Any) -> Any:
    if result.tag == 0:
        return folder(result.fields[0], state)

    else:
        return state


def Result_ForAll(predicate: Callable[[_A], bool], result: FSharpResult_2[Any, Any]) -> bool:
    if result.tag == 0:
        return predicate(result.fields[0])

    else:
        return True


def Result_Iterate(action: Callable[[_A], None], result: FSharpResult_2[Any, Any]) -> None:
    if result.tag == 0:
        action(result.fields[0])


def Result_ToArray(result: FSharpResult_2[Any, Any]) -> Array[Any]:
    if result.tag == 0:
        return [result.fields[0]]

    else:
        return []


def Result_ToList(result: FSharpResult_2[Any, Any]) -> FSharpList[Any]:
    if result.tag == 0:
        return singleton(result.fields[0])

    else:
        return empty()


def Result_ToOption(result: FSharpResult_2[Any, Any]) -> Any | None:
    if result.tag == 0:
        return some(result.fields[0])

    else:
        return None


def Result_ToValueOption(result: FSharpResult_2[Any, Any]) -> Any | None:
    if result.tag == 0:
        return some(result.fields[0])

    else:
        return None


__all__ = [
    "FSharpResult_2_reflection",
    "Result_Map",
    "Result_MapError",
    "Result_Bind",
    "Result_IsOk",
    "Result_IsError",
    "Result_Contains",
    "Result_Count",
    "Result_DefaultValue",
    "Result_DefaultWith",
    "Result_Exists",
    "Result_Fold",
    "Result_FoldBack",
    "Result_ForAll",
    "Result_Iterate",
    "Result_ToArray",
    "Result_ToList",
    "Result_ToOption",
    "Result_ToValueOption",
]



from __future__ import annotations
from collections.abc import Callable
from ..arctrl_core.arc_types import (ArcAssay, ArcStudy, ArcWorkflow, ArcRun, ArcInvestigation)
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ..fable_library.list import FSharpList
from ..fable_library.reflection import (TypeInfo, class_type)
from .arc import ARC
from .JsonIO.assay import (ARCtrl_ArcAssay__ArcAssay_fromJsonString_Static_Z721C83C5, ARCtrl_ArcAssay__ArcAssay_fromCompressedJsonString_Static_Z721C83C5, ARCtrl_ArcAssay__ArcAssay_fromISAJsonString_Static_Z721C83C5, ARCtrl_ArcAssay__ArcAssay_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_ArcAssay__ArcAssay_toJsonString_Static_71136F3F, ARCtrl_ArcAssay__ArcAssay_toCompressedJsonString_Static_71136F3F, ARCtrl_ArcAssay__ArcAssay_toISAJsonString_Static_Z3B036AA, ARCtrl_ArcAssay__ArcAssay_toROCrateJsonString_Static_5CABCA47)
from .JsonIO.datamap import (ARCtrl_DataMap__DataMap_fromJsonString_Static_Z721C83C5, ARCtrl_DataMap__DataMap_toJsonString_Static_71136F3F)
from .JsonIO.investigation import (ARCtrl_ArcInvestigation__ArcInvestigation_fromJsonString_Static_Z721C83C5, ARCtrl_ArcInvestigation__ArcInvestigation_fromCompressedJsonString_Static_Z721C83C5, ARCtrl_ArcInvestigation__ArcInvestigation_fromISAJsonString_Static_Z721C83C5, ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_ArcInvestigation__ArcInvestigation_toJsonString_Static_71136F3F, ARCtrl_ArcInvestigation__ArcInvestigation_toCompressedJsonString_Static_71136F3F, ARCtrl_ArcInvestigation__ArcInvestigation_toISAJsonString_Static_Z3B036AA, ARCtrl_ArcInvestigation__ArcInvestigation_toROCrateJsonString_Static_71136F3F)
from .JsonIO.ldobject import (ARCtrl_ROCrate_LDNode__LDNode_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_ROCrate_LDGraph__LDGraph_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_ROCrate_LDGraph__LDGraph_toROCrateJsonString_Static_71136F3F, ARCtrl_ROCrate_LDNode__LDNode_toROCrateJsonString_Static_71136F3F)
from .JsonIO.ontology_annotation import (ARCtrl_OntologyAnnotation__OntologyAnnotation_fromJsonString_Static_Z721C83C5, ARCtrl_OntologyAnnotation__OntologyAnnotation_fromISAJsonString_Static_Z721C83C5, ARCtrl_OntologyAnnotation__OntologyAnnotation_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_OntologyAnnotation__OntologyAnnotation_toJsonString_Static_71136F3F, ARCtrl_OntologyAnnotation__OntologyAnnotation_toISAJsonString_Static_71136F3F, ARCtrl_OntologyAnnotation__OntologyAnnotation_toROCrateJsonString_Static_71136F3F)
from .JsonIO.person import (ARCtrl_Person__Person_fromJsonString_Static_Z721C83C5, ARCtrl_Person__Person_fromISAJsonString_Static_Z721C83C5, ARCtrl_Person__Person_toJsonString_Static_71136F3F, ARCtrl_Person__Person_toISAJsonString_Static_Z3B036AA, ARCtrl_Person__Person_toROCrateJsonString_Static_71136F3F)
from .JsonIO.run import (ARCtrl_ArcRun__ArcRun_fromJsonString_Static_Z721C83C5, ARCtrl_ArcRun__ArcRun_fromCompressedJsonString_Static_Z721C83C5, ARCtrl_ArcRun__ArcRun_toJsonString_Static_71136F3F, ARCtrl_ArcRun__ArcRun_toCompressedJsonString_Static_71136F3F)
from .JsonIO.study import (ARCtrl_ArcStudy__ArcStudy_fromJsonString_Static_Z721C83C5, ARCtrl_ArcStudy__ArcStudy_fromCompressedJsonString_Static_Z721C83C5, ARCtrl_ArcStudy__ArcStudy_fromISAJsonString_Static_Z721C83C5, ARCtrl_ArcStudy__ArcStudy_fromROCrateJsonString_Static_Z721C83C5, ARCtrl_ArcStudy__ArcStudy_toJsonString_Static_71136F3F, ARCtrl_ArcStudy__ArcStudy_toCompressedJsonString_Static_71136F3F, ARCtrl_ArcStudy__ArcStudy_toISAJsonString_Static_Z3FD920F1, ARCtrl_ArcStudy__ArcStudy_toROCrateJsonString_Static_3BA23086)
from .JsonIO.workflow import (ARCtrl_ArcWorkflow__ArcWorkflow_fromJsonString_Static_Z721C83C5, ARCtrl_ArcWorkflow__ArcWorkflow_fromCompressedJsonString_Static_Z721C83C5, ARCtrl_ArcWorkflow__ArcWorkflow_toJsonString_Static_71136F3F, ARCtrl_ArcWorkflow__ArcWorkflow_toCompressedJsonString_Static_71136F3F)

def _expr4081() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.OntologyAnnotationJson", None, JsonHelper_OntologyAnnotationJson)


class JsonHelper_OntologyAnnotationJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> OntologyAnnotation:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_fromJsonString_Static_Z721C83C5(s)

    def from_isajson_string(self, s: str) -> OntologyAnnotation:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_fromISAJsonString_Static_Z721C83C5(s)

    def from_rocrate_json_string(self, s: str) -> OntologyAnnotation:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_json_string(self, oa: OntologyAnnotation, spaces: int | None=None) -> str:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_toJsonString_Static_71136F3F(spaces)(oa)

    def to_isajson_string(self, oa: OntologyAnnotation, spaces: int | None=None) -> str:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_toISAJsonString_Static_71136F3F(spaces)(oa)

    def to_rocrate_json_string(self, oa: OntologyAnnotation, spaces: int | None=None) -> str:
        return ARCtrl_OntologyAnnotation__OntologyAnnotation_toROCrateJsonString_Static_71136F3F(spaces)(oa)


JsonHelper_OntologyAnnotationJson_reflection = _expr4081

def JsonHelper_OntologyAnnotationJson__ctor(__unit: None=None) -> JsonHelper_OntologyAnnotationJson:
    return JsonHelper_OntologyAnnotationJson(__unit)


def _expr4082() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.PersonJson", None, JsonHelper_PersonJson)


class JsonHelper_PersonJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> Person:
        return ARCtrl_Person__Person_fromJsonString_Static_Z721C83C5(s)

    def from_isajson_string(self, s: str) -> Person:
        return ARCtrl_Person__Person_fromISAJsonString_Static_Z721C83C5(s)

    def from_rocrate_json_string(self, s: str) -> LDNode:
        return ARCtrl_ROCrate_LDNode__LDNode_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_json_string(self, person: Person, spaces: int | None=None) -> str:
        return ARCtrl_Person__Person_toJsonString_Static_71136F3F(spaces)(person)

    def to_isajson_string(self, person: Person, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
        return ARCtrl_Person__Person_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(person)

    def to_rocrate_json_string(self, person: Person, spaces: int | None=None) -> str:
        return ARCtrl_Person__Person_toROCrateJsonString_Static_71136F3F(spaces)(person)


JsonHelper_PersonJson_reflection = _expr4082

def JsonHelper_PersonJson__ctor(__unit: None=None) -> JsonHelper_PersonJson:
    return JsonHelper_PersonJson(__unit)


def _expr4083() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.DatamapJson", None, JsonHelper_DatamapJson)


class JsonHelper_DatamapJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> DataMap:
        return ARCtrl_DataMap__DataMap_fromJsonString_Static_Z721C83C5(s)

    def to_json_string(self, datamap: DataMap, spaces: int | None=None) -> str:
        return ARCtrl_DataMap__DataMap_toJsonString_Static_71136F3F(spaces)(datamap)


JsonHelper_DatamapJson_reflection = _expr4083

def JsonHelper_DatamapJson__ctor(__unit: None=None) -> JsonHelper_DatamapJson:
    return JsonHelper_DatamapJson(__unit)


def _expr4084() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.AssayJson", None, JsonHelper_AssayJson)


class JsonHelper_AssayJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromJsonString_Static_Z721C83C5(s)

    def from_compressed_json_string(self, s: str) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromCompressedJsonString_Static_Z721C83C5(s)

    def from_isajson_string(self, s: str) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromISAJsonString_Static_Z721C83C5(s)

    def from_rocrate_json_string(self, s: str) -> ArcAssay:
        return ARCtrl_ArcAssay__ArcAssay_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_json_string(self, assay: ArcAssay, spaces: int | None=None) -> str:
        return ARCtrl_ArcAssay__ArcAssay_toJsonString_Static_71136F3F(spaces)(assay)

    def to_compressed_json_string(self, assay: ArcAssay, spaces: int | None=None) -> str:
        return ARCtrl_ArcAssay__ArcAssay_toCompressedJsonString_Static_71136F3F(spaces)(assay)

    def to_isajson_string(self, assay: ArcAssay, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
        return ARCtrl_ArcAssay__ArcAssay_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(assay)

    def to_rocrate_json_string(self, assay: ArcAssay, study_name: str, spaces: int | None=None) -> str:
        return ARCtrl_ArcAssay__ArcAssay_toROCrateJsonString_Static_5CABCA47(study_name, spaces)(assay)


JsonHelper_AssayJson_reflection = _expr4084

def JsonHelper_AssayJson__ctor(__unit: None=None) -> JsonHelper_AssayJson:
    return JsonHelper_AssayJson(__unit)


def _expr4085() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.StudyJson", None, JsonHelper_StudyJson)


class JsonHelper_StudyJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> ArcStudy:
        return ARCtrl_ArcStudy__ArcStudy_fromJsonString_Static_Z721C83C5(s)

    def from_compressed_json_string(self, s: str) -> ArcStudy:
        return ARCtrl_ArcStudy__ArcStudy_fromCompressedJsonString_Static_Z721C83C5(s)

    def from_isajson_string(self, s: str) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
        return ARCtrl_ArcStudy__ArcStudy_fromISAJsonString_Static_Z721C83C5(s)

    def from_rocrate_json_string(self, s: str) -> tuple[ArcStudy, FSharpList[ArcAssay]]:
        return ARCtrl_ArcStudy__ArcStudy_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_json_string(self, study: ArcStudy, spaces: int | None=None) -> str:
        return ARCtrl_ArcStudy__ArcStudy_toJsonString_Static_71136F3F(spaces)(study)

    def to_compressed_json_string(self, study: ArcStudy, spaces: int | None=None) -> str:
        return ARCtrl_ArcStudy__ArcStudy_toCompressedJsonString_Static_71136F3F(spaces)(study)

    def to_isajson_string(self, study: ArcStudy, assays: FSharpList[ArcAssay] | None=None, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
        return ARCtrl_ArcStudy__ArcStudy_toISAJsonString_Static_Z3FD920F1(assays, spaces, use_idreferencing)(study)

    def to_rocrate_json_string(self, study: ArcStudy, assays: FSharpList[ArcAssay] | None=None, spaces: int | None=None) -> str:
        return ARCtrl_ArcStudy__ArcStudy_toROCrateJsonString_Static_3BA23086(assays, spaces)(study)


JsonHelper_StudyJson_reflection = _expr4085

def JsonHelper_StudyJson__ctor(__unit: None=None) -> JsonHelper_StudyJson:
    return JsonHelper_StudyJson(__unit)


def _expr4086() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.WorkflowJson", None, JsonHelper_WorkflowJson)


class JsonHelper_WorkflowJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> ArcWorkflow:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromJsonString_Static_Z721C83C5(s)

    def from_compressed_json_string(self, s: str) -> ArcWorkflow:
        return ARCtrl_ArcWorkflow__ArcWorkflow_fromCompressedJsonString_Static_Z721C83C5(s)

    def to_json_string(self, workflow: ArcWorkflow, spaces: int | None=None) -> str:
        return ARCtrl_ArcWorkflow__ArcWorkflow_toJsonString_Static_71136F3F(spaces)(workflow)

    def to_compressed_json_string(self, workflow: ArcWorkflow, spaces: int | None=None) -> str:
        return ARCtrl_ArcWorkflow__ArcWorkflow_toCompressedJsonString_Static_71136F3F(spaces)(workflow)


JsonHelper_WorkflowJson_reflection = _expr4086

def JsonHelper_WorkflowJson__ctor(__unit: None=None) -> JsonHelper_WorkflowJson:
    return JsonHelper_WorkflowJson(__unit)


def _expr4087() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.RunJson", None, JsonHelper_RunJson)


class JsonHelper_RunJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> ArcRun:
        return ARCtrl_ArcRun__ArcRun_fromJsonString_Static_Z721C83C5(s)

    def from_compressed_json_string(self, s: str) -> ArcRun:
        return ARCtrl_ArcRun__ArcRun_fromCompressedJsonString_Static_Z721C83C5(s)

    def to_json_string(self, run: ArcRun, spaces: int | None=None) -> str:
        return ARCtrl_ArcRun__ArcRun_toJsonString_Static_71136F3F(spaces)(run)

    def to_compressed_json_string(self, run: ArcRun, spaces: int | None=None) -> str:
        return ARCtrl_ArcRun__ArcRun_toCompressedJsonString_Static_71136F3F(spaces)(run)


JsonHelper_RunJson_reflection = _expr4087

def JsonHelper_RunJson__ctor(__unit: None=None) -> JsonHelper_RunJson:
    return JsonHelper_RunJson(__unit)


def _expr4088() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.InvestigationJson", None, JsonHelper_InvestigationJson)


class JsonHelper_InvestigationJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_json_string(self, s: str) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromJsonString_Static_Z721C83C5(s)

    def from_compressed_json_string(self, s: str) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromCompressedJsonString_Static_Z721C83C5(s)

    def from_isajson_string(self, s: str) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromISAJsonString_Static_Z721C83C5(s)

    def from_rocrate_json_string(self, s: str) -> ArcInvestigation:
        return ARCtrl_ArcInvestigation__ArcInvestigation_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_json_string(self, investigation: ArcInvestigation, spaces: int | None=None) -> str:
        return ARCtrl_ArcInvestigation__ArcInvestigation_toJsonString_Static_71136F3F(spaces)(investigation)

    def to_compressed_json_string(self, investigation: ArcInvestigation, spaces: int | None=None) -> str:
        return ARCtrl_ArcInvestigation__ArcInvestigation_toCompressedJsonString_Static_71136F3F(spaces)(investigation)

    def to_isajson_string(self, investigation: ArcInvestigation, spaces: int | None=None, use_idreferencing: bool | None=None) -> str:
        return ARCtrl_ArcInvestigation__ArcInvestigation_toISAJsonString_Static_Z3B036AA(spaces, use_idreferencing)(investigation)

    def to_rocrate_json_string(self, investigation: ArcInvestigation, spaces: int | None=None) -> str:
        return ARCtrl_ArcInvestigation__ArcInvestigation_toROCrateJsonString_Static_71136F3F(spaces)(investigation)


JsonHelper_InvestigationJson_reflection = _expr4088

def JsonHelper_InvestigationJson__ctor(__unit: None=None) -> JsonHelper_InvestigationJson:
    return JsonHelper_InvestigationJson(__unit)


def _expr4089() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.ARCJson", None, JsonHelper_ARCJson)


class JsonHelper_ARCJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_rocrate_json_string(self, s: str) -> ARC:
        return ARC.from_rocrate_json_string(s)

    def to_rocrate_json_string(self, spaces: int | None=None) -> Callable[[ARC], str]:
        return ARC.to_rocrate_json_string(spaces)


JsonHelper_ARCJson_reflection = _expr4089

def JsonHelper_ARCJson__ctor(__unit: None=None) -> JsonHelper_ARCJson:
    return JsonHelper_ARCJson(__unit)


def _expr4090() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.LDGraphJson", None, JsonHelper_LDGraphJson)


class JsonHelper_LDGraphJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_rocrate_json_string(self, s: str) -> LDGraph:
        return ARCtrl_ROCrate_LDGraph__LDGraph_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_rocrate_json_string(self, spaces: int | None=None) -> Callable[[LDGraph], str]:
        return ARCtrl_ROCrate_LDGraph__LDGraph_toROCrateJsonString_Static_71136F3F(spaces)


JsonHelper_LDGraphJson_reflection = _expr4090

def JsonHelper_LDGraphJson__ctor(__unit: None=None) -> JsonHelper_LDGraphJson:
    return JsonHelper_LDGraphJson(__unit)


def _expr4091() -> TypeInfo:
    return class_type("ARCtrl.JsonHelper.LDNodeJson", None, JsonHelper_LDNodeJson)


class JsonHelper_LDNodeJson:
    def __init__(self, __unit: None=None) -> None:
        pass

    def from_rocrate_json_string(self, s: str) -> LDNode:
        return ARCtrl_ROCrate_LDNode__LDNode_fromROCrateJsonString_Static_Z721C83C5(s)

    def to_rocrate_json_string(self, spaces: int | None=None) -> Callable[[LDNode], str]:
        return ARCtrl_ROCrate_LDNode__LDNode_toROCrateJsonString_Static_71136F3F(spaces)


JsonHelper_LDNodeJson_reflection = _expr4091

def JsonHelper_LDNodeJson__ctor(__unit: None=None) -> JsonHelper_LDNodeJson:
    return JsonHelper_LDNodeJson(__unit)


def _expr4092() -> TypeInfo:
    return class_type("ARCtrl.JsonController", None, JsonController)


class JsonController:
    @staticmethod
    def OntologyAnnotation() -> JsonHelper_OntologyAnnotationJson:
        return JsonHelper_OntologyAnnotationJson()

    @staticmethod
    def Person() -> JsonHelper_PersonJson:
        return JsonHelper_PersonJson()

    @staticmethod
    def Datamap() -> JsonHelper_DatamapJson:
        return JsonHelper_DatamapJson()

    @staticmethod
    def Assay() -> JsonHelper_AssayJson:
        return JsonHelper_AssayJson()

    @staticmethod
    def Study() -> JsonHelper_StudyJson:
        return JsonHelper_StudyJson()

    @staticmethod
    def Workflow() -> JsonHelper_WorkflowJson:
        return JsonHelper_WorkflowJson()

    @staticmethod
    def Run() -> JsonHelper_RunJson:
        return JsonHelper_RunJson()

    @staticmethod
    def Investigation() -> JsonHelper_InvestigationJson:
        return JsonHelper_InvestigationJson()

    @staticmethod
    def ARC() -> JsonHelper_ARCJson:
        return JsonHelper_ARCJson()

    @staticmethod
    def LDGraph() -> JsonHelper_LDGraphJson:
        return JsonHelper_LDGraphJson()

    @staticmethod
    def LDNode() -> JsonHelper_LDNodeJson:
        return JsonHelper_LDNodeJson()


JsonController_reflection = _expr4092

__all__ = ["JsonHelper_OntologyAnnotationJson_reflection", "JsonHelper_PersonJson_reflection", "JsonHelper_DatamapJson_reflection", "JsonHelper_AssayJson_reflection", "JsonHelper_StudyJson_reflection", "JsonHelper_WorkflowJson_reflection", "JsonHelper_RunJson_reflection", "JsonHelper_InvestigationJson_reflection", "JsonHelper_ARCJson_reflection", "JsonHelper_LDGraphJson_reflection", "JsonHelper_LDNodeJson_reflection", "JsonController_reflection"]




from __future__ import annotations
from typing import Any
from ...arctrl_core.arc_types import ArcWorkflow
from ...arctrl_core.comment import Comment
from ...arctrl_core.data_map import DataMap
from ...arctrl_core.Helper.collections_ import (ResizeArray_map, ResizeArray_singleton, ResizeArray_choose, Option_fromSeq, Option_fromValueWithDefault)
from ...arctrl_core.Helper.identifier import Workflow_cwlFileNameFromIdentifier
from ...arctrl_core.ontology_annotation import OntologyAnnotation
from ...arctrl_core.person import Person
from ...arctrl_core.Process.component import (Component, Component_create_Z2F0B38C7)
from ...arctrl_core.value import Value as Value_3
from ...arctrl_cwl.cwlprocessing_unit import CWLProcessingUnit
from ...arctrl_cwl.cwltypes import CWLType
from ...arctrl_cwl.decode import Decode_cwlTypeStringMatcher
from ...arctrl_cwl.encode import (write_yaml, encode_cwltype)
from ...arctrl_cwl.inputs import (InputBinding, CWLInput, InputBinding_create_ZAC0108A)
from ...arctrl_cwl.outputs import (OutputBinding, CWLOutput, OutputBinding_create_6DFDD678)
from ...arctrl_cwl.tool_description import CWLToolDescription
from ...arctrl_cwl.workflow_description import CWLWorkflowDescription
from ...arctrl_cwl.workflow_steps import (WorkflowStep, StepOutput)
from ...arctrl_file_system.file_system import FileSystem
from ...arctrl_rocrate.ldcontext import LDContext
from ...arctrl_rocrate.ldobject import (LDNode, LDGraph)
from ...arctrl_rocrate.LDTypes.computational_workflow import LDComputationalWorkflow
from ...arctrl_rocrate.LDTypes.computer_language import LDComputerLanguage
from ...arctrl_rocrate.LDTypes.dataset import LDDataset
from ...arctrl_rocrate.LDTypes.formal_parameter import LDFormalParameter
from ...arctrl_rocrate.LDTypes.lab_protocol import LDLabProtocol
from ...arctrl_rocrate.LDTypes.property_value import LDPropertyValue
from ...arctrl_rocrate.LDTypes.workflow_protocol import LDWorkflowProtocol
from ...fable_library.option import (default_arg, bind, map, value as value_2)
from ...fable_library.reflection import (TypeInfo, class_type)
from ...fable_library.seq import (try_pick, exactly_one, find)
from ...fable_library.string_ import (to_fail, printf)
from ...fable_library.types import (Array, to_string)
from ...fable_library.util import int32_to_string
from ...yamlicious.decode import (object, IGetters)
from ...yamlicious.yamlicious_types import YAMLElement
from .basic import (BaseTypes_decomposePropertyValue_Z2F770004, BaseTypes_composeDefinedTerm_ZDED3A0F, BaseTypes_composeComment_Z13201A7E, BaseTypes_decomposeDefinedTerm_Z2F770004, BaseTypes_decomposeComment_Z2F770004)
from .datamap import (DatamapConversion_composeFragmentDescriptors_Z8923FA3, DatamapConversion_decomposeFragmentDescriptors_Z6E59645F)
from .person import (PersonConversion_composePerson_Z64D846DC, PersonConversion_decomposePerson_Z6839B9E8)

def _expr3957() -> TypeInfo:
    return class_type("ARCtrl.Conversion.WorkflowConversion", None, WorkflowConversion)


class WorkflowConversion:
    ...

WorkflowConversion_reflection = _expr3957

def WorkflowConversion_composeAdditionalType_Z5C31A0F9(t: CWLType) -> str:
    s: str = write_yaml(encode_cwltype(t))
    return s.strip()


def WorkflowConversion_decomposeAdditionalType_Z721C83C5(t: str) -> CWLType:
    def _arrow3958(get: IGetters, t: Any=t) -> tuple[CWLType, bool]:
        return Decode_cwlTypeStringMatcher(t, get)

    return object(_arrow3958, YAMLElement(4, "dawd"))[0]


def WorkflowConversion_composeFormalParamInputIdentifiers(prefix: str | None=None, position: int | None=None) -> Array[LDNode] | None:
    if prefix is None:
        if position is None:
            return None

        else: 
            po_1: int = position or 0
            return [LDPropertyValue.create_position(po_1)]


    elif position is None:
        pr_1: str = prefix
        return [LDPropertyValue.create_prefix(pr_1)]

    else: 
        po: int = position or 0
        pr: str = prefix
        return [LDPropertyValue.create_position(po), LDPropertyValue.create_prefix(pr)]



def WorkflowConversion_composeFormalParameterFromInput_Z7A02DED7(inp: CWLInput, workflow_name: str | None=None, run_name: str | None=None) -> LDNode:
    additional_type: str
    match_value: CWLType | None = inp.Type_
    if match_value is None:
        raise Exception("Input must have a type")

    else: 
        additional_type = WorkflowConversion_composeAdditionalType_Z5C31A0F9(match_value)

    value_required: bool = not default_arg(inp.Optional, False)
    id: str = LDFormalParameter.gen_id(inp.Name, workflow_name, run_name)
    def binder(ib: InputBinding, inp: Any=inp, workflow_name: Any=workflow_name, run_name: Any=run_name) -> Array[LDNode] | None:
        return WorkflowConversion_composeFormalParamInputIdentifiers(ib.Prefix, ib.Position)

    identifiers: Array[LDNode] | None = bind(binder, inp.InputBinding)
    return LDFormalParameter.create(additional_type, id, None, inp.Name, None, None, None, None, value_required, identifiers)


def WorkflowConversion_decomposeInputBindings_1AAAE9A5(identifiers: Array[LDNode], context: LDContext | None=None) -> InputBinding | None:
    def _arrow3959(n: LDNode, identifiers: Any=identifiers, context: Any=context) -> str | None:
        return LDPropertyValue.try_get_as_prefix(n, context)

    prefix: str | None = try_pick(_arrow3959, identifiers)
    def _arrow3960(n_1: LDNode, identifiers: Any=identifiers, context: Any=context) -> int | None:
        return LDPropertyValue.try_get_as_position(n_1, context)

    position: int | None = try_pick(_arrow3960, identifiers)
    (pattern_matching_result,) = (None,)
    if prefix is None:
        if position is None:
            pattern_matching_result = 0

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        return None

    elif pattern_matching_result == 1:
        return InputBinding_create_ZAC0108A(prefix, position)



def WorkflowConversion_decomposeInputFromFormalParameter_2664C4B8(inp: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLInput:
    t_1: CWLType = WorkflowConversion_decomposeAdditionalType_Z721C83C5(exactly_one(inp.AdditionalType))
    binding: InputBinding | None = WorkflowConversion_decomposeInputBindings_1AAAE9A5(LDFormalParameter.get_identifiers(inp, graph, context), context)
    optional: bool | None
    match_value: bool | None = LDFormalParameter.try_get_value_required_as_boolean(inp, context)
    (pattern_matching_result,) = (None,)
    if match_value is not None:
        if match_value:
            pattern_matching_result = 0

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        optional = None

    elif pattern_matching_result == 1:
        optional = True

    return CWLInput(LDFormalParameter.get_name_as_string(inp, context), t_1, binding, optional)


def WorkflowConversion_composeFormalParameterOutputIdentifiers_6DFDD678(glob: str | None=None) -> Array[LDNode] | None:
    if glob is None:
        return None

    else: 
        g: str = glob
        return [LDPropertyValue.create_glob(g)]



def WorkflowConversion_composeFormalParameterFromOutput_Z25C7EC80(out: CWLOutput, workflow_name: str | None=None, run_name: str | None=None) -> LDNode:
    additional_type: str
    match_value: CWLType | None = out.Type_
    if match_value is None:
        raise Exception("Output must have a type")

    else: 
        additional_type = WorkflowConversion_composeAdditionalType_Z5C31A0F9(match_value)

    id: str = LDFormalParameter.gen_id(out.Name, workflow_name, run_name)
    def binder(ob: OutputBinding, out: Any=out, workflow_name: Any=workflow_name, run_name: Any=run_name) -> Array[LDNode] | None:
        return WorkflowConversion_composeFormalParameterOutputIdentifiers_6DFDD678(ob.Glob)

    identifiers: Array[LDNode] | None = bind(binder, out.OutputBinding)
    return LDFormalParameter.create(additional_type, id, None, out.Name, None, None, None, None, True, identifiers)


def WorkflowConversion_decomposeOutputBindings_1AAAE9A5(identifiers: Array[LDNode], context: LDContext | None=None) -> OutputBinding:
    def _arrow3962(n: LDNode, identifiers: Any=identifiers, context: Any=context) -> str | None:
        return LDPropertyValue.try_get_as_glob(n, context)

    return OutputBinding_create_6DFDD678(try_pick(_arrow3962, identifiers))


def WorkflowConversion_decomposeOutputFromFormalParameter_2664C4B8(inp: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLOutput:
    t_1: CWLType = WorkflowConversion_decomposeAdditionalType_Z721C83C5(exactly_one(inp.AdditionalType))
    binding: OutputBinding = WorkflowConversion_decomposeOutputBindings_1AAAE9A5(LDFormalParameter.get_identifiers(inp, graph, context), context)
    def _arrow3963(value: bool, inp: Any=inp, context: Any=context, graph: Any=graph) -> bool:
        return not value

    optional: bool | None = map(_arrow3963, LDFormalParameter.try_get_value_required_as_boolean(inp, context))
    return CWLOutput(LDFormalParameter.get_name_as_string(inp, context), t_1, binding)


def WorkflowConversion_composeComputationalTool_Z685B8F25(tool: Component) -> LDNode:
    pattern_input: tuple[str, str | None]
    match_value: OntologyAnnotation | None = tool.ComponentType
    (pattern_matching_result, c_1) = (None, None)
    if match_value is not None:
        def _arrow3964(__unit: None=None, tool: Any=tool) -> bool:
            c: OntologyAnnotation = match_value
            return c.Name is not None

        if _arrow3964():
            pattern_matching_result = 0
            c_1 = match_value

        else: 
            pattern_matching_result = 1


    else: 
        pattern_matching_result = 1

    if pattern_matching_result == 0:
        pattern_input = (value_2(c_1.Name), c_1.TermAccessionNumber)

    elif pattern_matching_result == 1:
        raise Exception("Component must have a type")

    pattern_input_1: tuple[str | None, str | None]
    match_value_1: Value_3 | None = tool.ComponentValue
    def _arrow3965(__unit: None=None, tool: Any=tool) -> tuple[str | None, str | None]:
        f: float = match_value_1.fields[0]
        return (to_string(f), None)

    def _arrow3966(__unit: None=None, tool: Any=tool) -> tuple[str | None, str | None]:
        i: int = match_value_1.fields[0] or 0
        return (int32_to_string(i), None)

    def _arrow3967(__unit: None=None, tool: Any=tool) -> tuple[str | None, str | None]:
        oa: OntologyAnnotation = match_value_1.fields[0]
        return (oa.Name, oa.TermAccessionNumber)

    def _arrow3968(__unit: None=None, tool: Any=tool) -> tuple[str | None, str | None]:
        s: str = match_value_1.fields[0]
        return (s, None)

    pattern_input_1 = ((None, None)) if (match_value_1 is None) else (_arrow3965() if (match_value_1.tag == 2) else (_arrow3966() if (match_value_1.tag == 1) else (_arrow3967() if (match_value_1.tag == 0) else _arrow3968())))
    pattern_input_2: tuple[str | None, str | None]
    match_value_2: OntologyAnnotation | None = tool.ComponentUnit
    if match_value_2 is None:
        pattern_input_2 = (None, None)

    else: 
        oa_1: OntologyAnnotation = match_value_2
        pattern_input_2 = (oa_1.Name, oa_1.TermAccessionNumber)

    return LDPropertyValue.create_component(pattern_input[0], pattern_input_1[0], None, pattern_input[1], pattern_input_2[1], pattern_input_2[0], pattern_input_1[1])


def WorkflowConversion_decomposeComputationalTool_Z6839B9E8(tool: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> Component:
    pattern_input: tuple[OntologyAnnotation, Value_3 | None, OntologyAnnotation | None] = BaseTypes_decomposePropertyValue_Z2F770004(tool, context)
    return Component_create_Z2F0B38C7(pattern_input[1], pattern_input[2], pattern_input[0])


def WorkflowConversion_getInputParametersFromProcessingUnit_30922B92(pu: CWLProcessingUnit) -> Array[CWLInput]:
    if pu.tag == 1:
        def f_1(i_1: CWLInput, pu: Any=pu) -> CWLInput:
            return i_1

        return ResizeArray_map(f_1, pu.fields[0].Inputs)

    else: 
        match_value: Array[CWLInput] | None = pu.fields[0].Inputs
        if match_value is None:
            return []

        else: 
            def f(i: CWLInput, pu: Any=pu) -> CWLInput:
                return i

            return ResizeArray_map(f, match_value)




def WorkflowConversion_get_toolDescriptionTypeName(__unit: None=None) -> str:
    return "ToolDescription"


def WorkflowConversion_get_workflowDescriptionTypeName(__unit: None=None) -> str:
    return "WorkflowDescription"


def WorkflowConversion_composeWorkflowProtocolFromToolDescription_7B8357DA(file_path: str, workflow: CWLToolDescription, workflow_name: str | None=None, run_name: str | None=None) -> LDNode:
    def mapping(a: Array[CWLInput], file_path: Any=file_path, workflow: Any=workflow, workflow_name: Any=workflow_name, run_name: Any=run_name) -> Array[LDNode]:
        def f(i: CWLInput, a: Any=a) -> LDNode:
            return WorkflowConversion_composeFormalParameterFromInput_Z7A02DED7(i, workflow_name, run_name)

        return ResizeArray_map(f, a)

    inputs: Array[LDNode] | None = map(mapping, workflow.Inputs)
    def f_1(o: CWLOutput, file_path: Any=file_path, workflow: Any=workflow, workflow_name: Any=workflow_name, run_name: Any=run_name) -> LDNode:
        return WorkflowConversion_composeFormalParameterFromOutput_Z25C7EC80(o, workflow_name, run_name)

    outputs: Array[LDNode] = ResizeArray_map(f_1, workflow.Outputs)
    return LDWorkflowProtocol.create(file_path, inputs, outputs, None, None, None, None, ResizeArray_singleton(LDComputerLanguage.create_cwl()), None, None, None, None, None, None, None, None, [WorkflowConversion_get_toolDescriptionTypeName()])


def WorkflowConversion_decomposeWorkflowProtocolToToolDescription_2664C4B8(protocol: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLToolDescription:
    def f(i: LDNode, protocol: Any=protocol, context: Any=context, graph: Any=graph) -> CWLInput:
        return WorkflowConversion_decomposeInputFromFormalParameter_2664C4B8(i, context, graph)

    inputs: Array[CWLInput] = ResizeArray_map(f, LDComputationalWorkflow.get_inputs_as_formal_parameters(protocol, graph, context))
    def f_1(o: LDNode, protocol: Any=protocol, context: Any=context, graph: Any=graph) -> CWLOutput:
        return WorkflowConversion_decomposeOutputFromFormalParameter_2664C4B8(o, context, graph)

    return CWLToolDescription(ResizeArray_map(f_1, LDComputationalWorkflow.get_outputs_as_formal_parameter(protocol, graph, context)), None, None, None, None, inputs)


def WorkflowConversion_composeWorkflowStep_Z35AB9E5D(step: WorkflowStep, workflow_id: Any) -> LDNode:
    id: str = ((("WorkflowStep_" + str(workflow_id)) + "_") + step.Id) + ""
    cw: LDNode = LDComputationalWorkflow.create(id, None, None, None, None, None, step.Id)
    LDDataset.set_identifier_as_string(cw, step.Run)
    return cw


def WorkflowConversion_composeWorkflowProtocolFromWorkflow_Z6FBB852F(file_path: str, workflow: CWLWorkflowDescription, workflow_name: str | None=None, run_name: str | None=None) -> LDNode:
    def f(i: CWLInput, file_path: Any=file_path, workflow: Any=workflow, workflow_name: Any=workflow_name, run_name: Any=run_name) -> LDNode:
        return WorkflowConversion_composeFormalParameterFromInput_Z7A02DED7(i, workflow_name, run_name)

    inputs: Array[LDNode] = ResizeArray_map(f, workflow.Inputs)
    def f_1(o: CWLOutput, file_path: Any=file_path, workflow: Any=workflow, workflow_name: Any=workflow_name, run_name: Any=run_name) -> LDNode:
        return WorkflowConversion_composeFormalParameterFromOutput_Z25C7EC80(o, workflow_name, run_name)

    outputs: Array[LDNode] = ResizeArray_map(f_1, workflow.Outputs)
    def f_2(s: WorkflowStep, file_path: Any=file_path, workflow: Any=workflow, workflow_name: Any=workflow_name, run_name: Any=run_name) -> LDNode:
        return WorkflowConversion_composeWorkflowStep_Z35AB9E5D(s, file_path)

    steps: Array[LDNode] = ResizeArray_map(f_2, workflow.Steps)
    return LDWorkflowProtocol.create(file_path, inputs, outputs, None, None, None, None, ResizeArray_singleton(LDComputerLanguage.create_cwl()), None, None, None, None, steps, None, None, None, [WorkflowConversion_get_workflowDescriptionTypeName()])


def WorkflowConversion_decomposeWorkflowStep_2664C4B8(step: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> WorkflowStep:
    name: str = LDComputationalWorkflow.get_name_as_string(step, context)
    run: str = LDDataset.get_identifier_as_string(step, context)
    return WorkflowStep(name, [], StepOutput.create([]), run)


def WorkflowConversion_decomposeWorkflowProtocolToWorkflow_2664C4B8(protocol: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLWorkflowDescription:
    def f(i: LDNode, protocol: Any=protocol, context: Any=context, graph: Any=graph) -> CWLInput:
        return WorkflowConversion_decomposeInputFromFormalParameter_2664C4B8(i, context, graph)

    inputs: Array[CWLInput] = ResizeArray_map(f, LDComputationalWorkflow.get_inputs_as_formal_parameters(protocol, graph, context))
    def f_1(o: LDNode, protocol: Any=protocol, context: Any=context, graph: Any=graph) -> CWLOutput:
        return WorkflowConversion_decomposeOutputFromFormalParameter_2664C4B8(o, context, graph)

    outputs: Array[CWLOutput] = ResizeArray_map(f_1, LDComputationalWorkflow.get_outputs_as_formal_parameter(protocol, graph, context))
    def f_2(s: LDNode, protocol: Any=protocol, context: Any=context, graph: Any=graph) -> WorkflowStep:
        return WorkflowConversion_decomposeWorkflowStep_2664C4B8(s, context, graph)

    return CWLWorkflowDescription(ResizeArray_map(f_2, LDComputationalWorkflow.get_has_part(protocol, graph, context)), inputs, outputs)


def WorkflowConversion_composeWorkflowProtocolFromProcessingUnit_Z3099C0F7(file_path: str, pu: CWLProcessingUnit, workflow_name: str | None=None, run_name: str | None=None) -> LDNode:
    if pu.tag == 1:
        return WorkflowConversion_composeWorkflowProtocolFromWorkflow_Z6FBB852F(file_path, pu.fields[0], workflow_name, run_name)

    else: 
        return WorkflowConversion_composeWorkflowProtocolFromToolDescription_7B8357DA(file_path, pu.fields[0], workflow_name, run_name)



def WorkflowConversion_decomposeWorkflowProtocolToProcessingUnit_2664C4B8(protocol: LDNode, context: LDContext | None=None, graph: LDGraph | None=None) -> CWLProcessingUnit:
    if LDComputationalWorkflow.get_additional_type_as_string(protocol, context) == WorkflowConversion_get_workflowDescriptionTypeName():
        return CWLProcessingUnit(1, WorkflowConversion_decomposeWorkflowProtocolToWorkflow_2664C4B8(protocol, context, graph))

    else: 
        return CWLProcessingUnit(0, WorkflowConversion_decomposeWorkflowProtocolToToolDescription_2664C4B8(protocol, context, graph))



def WorkflowConversion_composeWorkflow_42450E6E(workflow: ArcWorkflow, fs: FileSystem | None=None) -> LDNode:
    workflow_protocol: LDNode
    workflow_file_path: str = Workflow_cwlFileNameFromIdentifier(workflow.Identifier)
    match_value: CWLProcessingUnit | None = workflow.CWLDescription
    if match_value is None:
        arg: str = workflow.Identifier
        workflow_protocol = to_fail(printf("Workflow %s must have a CWL description"))(arg)

    else: 
        workflow_protocol = WorkflowConversion_composeWorkflowProtocolFromProcessingUnit_Z3099C0F7(workflow_file_path, match_value, workflow.Identifier)

    if workflow.Version is not None:
        LDLabProtocol.set_version_as_string(workflow_protocol, value_2(workflow.Version))

    if workflow.URI is not None:
        LDLabProtocol.set_url(workflow_protocol, value_2(workflow.URI))

    if workflow.WorkflowType is not None:
        intended_use: LDNode = BaseTypes_composeDefinedTerm_ZDED3A0F(value_2(workflow.WorkflowType))
        LDLabProtocol.set_intended_use_as_defined_term(workflow_protocol, intended_use)

    def f(df: LDNode, workflow: Any=workflow, fs: Any=fs) -> LDNode | None:
        return LDPropertyValue.try_get_subject_of(df)

    def mapping(datamap: DataMap, workflow: Any=workflow, fs: Any=fs) -> Array[LDNode]:
        return DatamapConversion_composeFragmentDescriptors_Z8923FA3(datamap)

    data_files: Array[LDNode] = ResizeArray_choose(f, default_arg(map(mapping, workflow.DataMap), []))
    if len(workflow.Components) > 0:
        def f_1(tool: Component, workflow: Any=workflow, fs: Any=fs) -> LDNode:
            return WorkflowConversion_composeComputationalTool_Z685B8F25(tool)

        software_tools: Array[LDNode] = ResizeArray_map(f_1, workflow.Components)
        LDLabProtocol.set_computational_tools(workflow_protocol, software_tools)

    def f_2(c: Person, workflow: Any=workflow, fs: Any=fs) -> LDNode:
        return PersonConversion_composePerson_Z64D846DC(c)

    creators: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_2, workflow.Contacts))
    has_parts: Array[LDNode] | None = Option_fromSeq(data_files)
    def f_3(c_1: Comment, workflow: Any=workflow, fs: Any=fs) -> LDNode:
        return BaseTypes_composeComment_Z13201A7E(c_1)

    comments: Array[LDNode] | None = Option_fromSeq(ResizeArray_map(f_3, workflow.Comments))
    return LDDataset.create_arcworkflow(workflow.Identifier, ResizeArray_singleton(workflow_protocol), None, workflow.Title, workflow.Description, creators, has_parts, comments)


def WorkflowConversion_decomposeWorkflow_Z6839B9E8(workflow: LDNode, graph: LDGraph | None=None, context: LDContext | None=None) -> ArcWorkflow:
    def predicate(me: LDNode, workflow: Any=workflow, graph: Any=graph, context: Any=context) -> bool:
        return LDComputationalWorkflow.validate(me, context)

    workflow_protocol: LDNode = find(predicate, LDDataset.get_main_entities(workflow, graph, context))
    cwl_description: CWLProcessingUnit = WorkflowConversion_decomposeWorkflowProtocolToProcessingUnit_2664C4B8(workflow_protocol, context, graph)
    version: str | None = LDLabProtocol.try_get_version_as_string(workflow_protocol, context)
    uri: str | None = LDLabProtocol.try_get_url(workflow_protocol, context)
    def mapping(iu: LDNode, workflow: Any=workflow, graph: Any=graph, context: Any=context) -> OntologyAnnotation:
        return BaseTypes_decomposeDefinedTerm_Z2F770004(iu, context)

    workflow_type: OntologyAnnotation | None = map(mapping, LDLabProtocol.try_get_intended_use_as_defined_term(workflow_protocol, graph, context))
    def f(ct: LDNode, workflow: Any=workflow, graph: Any=graph, context: Any=context) -> Component:
        return WorkflowConversion_decomposeComputationalTool_Z6839B9E8(ct, graph, context)

    components: Array[Component] = ResizeArray_map(f, LDLabProtocol.get_computational_tools(workflow_protocol, graph, context))
    def f_1(c: LDNode, workflow: Any=workflow, graph: Any=graph, context: Any=context) -> Person:
        return PersonConversion_decomposePerson_Z6839B9E8(c, graph, context)

    contacts: Array[Person] = ResizeArray_map(f_1, LDDataset.get_creators(workflow, graph, context))
    def f_2(c_1: LDNode, workflow: Any=workflow, graph: Any=graph, context: Any=context) -> Comment:
        return BaseTypes_decomposeComment_Z2F770004(c_1, context)

    comments: Array[Comment] = ResizeArray_map(f_2, LDDataset.get_comments(workflow, graph, context))
    data_map: DataMap | None
    v: DataMap = DatamapConversion_decomposeFragmentDescriptors_Z6E59645F(LDDataset.get_variable_measured_as_fragment_descriptors(workflow, graph, context), graph, context)
    data_map = Option_fromValueWithDefault(DataMap.init(), v)
    return ArcWorkflow.create(LDDataset.get_identifier_as_string(workflow, context), LDDataset.try_get_name_as_string(workflow, context), LDDataset.try_get_description_as_string(workflow, context), workflow_type, uri, version, None, None, components, data_map, contacts, cwl_description, comments)


__all__ = ["WorkflowConversion_reflection", "WorkflowConversion_composeAdditionalType_Z5C31A0F9", "WorkflowConversion_decomposeAdditionalType_Z721C83C5", "WorkflowConversion_composeFormalParamInputIdentifiers", "WorkflowConversion_composeFormalParameterFromInput_Z7A02DED7", "WorkflowConversion_decomposeInputBindings_1AAAE9A5", "WorkflowConversion_decomposeInputFromFormalParameter_2664C4B8", "WorkflowConversion_composeFormalParameterOutputIdentifiers_6DFDD678", "WorkflowConversion_composeFormalParameterFromOutput_Z25C7EC80", "WorkflowConversion_decomposeOutputBindings_1AAAE9A5", "WorkflowConversion_decomposeOutputFromFormalParameter_2664C4B8", "WorkflowConversion_composeComputationalTool_Z685B8F25", "WorkflowConversion_decomposeComputationalTool_Z6839B9E8", "WorkflowConversion_getInputParametersFromProcessingUnit_30922B92", "WorkflowConversion_get_toolDescriptionTypeName", "WorkflowConversion_get_workflowDescriptionTypeName", "WorkflowConversion_composeWorkflowProtocolFromToolDescription_7B8357DA", "WorkflowConversion_decomposeWorkflowProtocolToToolDescription_2664C4B8", "WorkflowConversion_composeWorkflowStep_Z35AB9E5D", "WorkflowConversion_composeWorkflowProtocolFromWorkflow_Z6FBB852F", "WorkflowConversion_decomposeWorkflowStep_2664C4B8", "WorkflowConversion_decomposeWorkflowProtocolToWorkflow_2664C4B8", "WorkflowConversion_composeWorkflowProtocolFromProcessingUnit_Z3099C0F7", "WorkflowConversion_decomposeWorkflowProtocolToProcessingUnit_2664C4B8", "WorkflowConversion_composeWorkflow_42450E6E", "WorkflowConversion_decomposeWorkflow_Z6839B9E8"]




from __future__ import annotations
from collections.abc import Callable
from typing import (Any, TypeVar)
from ..arctrl_core.arc_types import ArcAssay
from ..arctrl_core.comment import Comment
from ..arctrl_core.conversion import (ARCtrl_ArcTables__ArcTables_GetProcesses, ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D, JsonTypes_composeTechnologyPlatform, JsonTypes_decomposeTechnologyPlatform)
from ..arctrl_core.data import Data
from ..arctrl_core.data_map import DataMap
from ..arctrl_core.Helper.collections_ import (Option_fromValueWithDefault, ResizeArray_filter)
from ..arctrl_core.Helper.identifier import (Assay_fileNameFromIdentifier, create_missing_identifier, Assay_tryIdentifierFromFileName)
from ..arctrl_core.ontology_annotation import OntologyAnnotation
from ..arctrl_core.person import Person
from ..arctrl_core.Process.material_attribute import MaterialAttribute
from ..arctrl_core.Process.process import Process
from ..arctrl_core.Process.process_sequence import (get_data, get_units, get_characteristics)
from ..arctrl_core.Table.arc_table import ArcTable
from ..arctrl_core.Table.arc_tables import ArcTables
from ..arctrl_core.Table.composite_cell import CompositeCell
from ..fable_library.list import (choose, of_array, FSharpList, singleton, empty)
from ..fable_library.option import (map, default_arg, value as value_9, bind)
from ..fable_library.seq import (map as map_1, to_list, delay, append, empty as empty_1, singleton as singleton_1, try_pick, length)
from ..fable_library.string_ import replace
from ..fable_library.types import Array
from ..fable_library.util import IEnumerable_1
from ..thoth_json_core.decode import (object, IRequiredGetter, string, IOptionalGetter, resize_array, IGetters, list_1 as list_1_2, map as map_2)
from ..thoth_json_core.encode import list_1 as list_1_1
from ..thoth_json_core.types import (IEncodable, IEncoderHelpers_1, Decoder_1)
from .comment import (encoder as encoder_7, decoder as decoder_4, ROCrate_encoder as ROCrate_encoder_4, ROCrate_decoder as ROCrate_decoder_3, ISAJson_encoder as ISAJson_encoder_3, ISAJson_decoder as ISAJson_decoder_2)
from .context.rocrate.isa_assay_context import context_jsonvalue
from .data import (ROCrate_encoder as ROCrate_encoder_2, ISAJson_encoder as ISAJson_encoder_1)
from .DataMap.data_map import (encoder as encoder_4, decoder as decoder_2, encoder_compressed as encoder_compressed_2, decoder_compressed as decoder_compressed_2)
from .decode import Decode_objectNoAdditionalProperties
from .encode import (try_include, try_include_seq, try_include_list)
from .idtable import encode
from .ontology_annotation import (OntologyAnnotation_encoder, OntologyAnnotation_decoder, OntologyAnnotation_ROCrate_encoderPropertyValue, OntologyAnnotation_ROCrate_encoderDefinedTerm, OntologyAnnotation_ROCrate_decoderPropertyValue, OntologyAnnotation_ROCrate_decoderDefinedTerm, OntologyAnnotation_ISAJson_encoder, OntologyAnnotation_ISAJson_decoder)
from .person import (encoder as encoder_6, decoder as decoder_3, ROCrate_encoder as ROCrate_encoder_1, ROCrate_decoder as ROCrate_decoder_2)
from .Process.assay_materials import encoder as encoder_9
from .Process.material_attribute import encoder as encoder_8
from .Process.process import (ROCrate_encoder as ROCrate_encoder_3, ROCrate_decoder as ROCrate_decoder_1, ISAJson_encoder as ISAJson_encoder_2, ISAJson_decoder as ISAJson_decoder_1)
from .Table.arc_table import (encoder as encoder_5, decoder as decoder_1, encoder_compressed as encoder_compressed_1, decoder_compressed as decoder_compressed_1)

__A_ = TypeVar("__A_")

def encoder(assay: ArcAssay) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], assay: Any=assay) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2885(__unit: None=None, assay: Any=assay) -> IEncodable:
        value: str = assay.Identifier
        class ObjectExpr2884(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2884()

    def _arrow2887(value_1: str, assay: Any=assay) -> IEncodable:
        class ObjectExpr2886(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr2886()

    def _arrow2889(value_3: str, assay: Any=assay) -> IEncodable:
        class ObjectExpr2888(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr2888()

    def _arrow2890(oa: OntologyAnnotation, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2891(oa_1: OntologyAnnotation, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2892(oa_2: OntologyAnnotation, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa_2)

    def _arrow2893(dm: DataMap, assay: Any=assay) -> IEncodable:
        return encoder_4(dm)

    def _arrow2894(table: ArcTable, assay: Any=assay) -> IEncodable:
        return encoder_5(table)

    def _arrow2895(person: Person, assay: Any=assay) -> IEncodable:
        return encoder_6(person)

    def _arrow2896(comment: Comment, assay: Any=assay) -> IEncodable:
        return encoder_7(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow2885()), try_include("Title", _arrow2887, assay.Title), try_include("Description", _arrow2889, assay.Description), try_include("MeasurementType", _arrow2890, assay.MeasurementType), try_include("TechnologyType", _arrow2891, assay.TechnologyType), try_include("TechnologyPlatform", _arrow2892, assay.TechnologyPlatform), try_include("DataMap", _arrow2893, assay.DataMap), try_include_seq("Tables", _arrow2894, assay.Tables), try_include_seq("Performers", _arrow2895, assay.Performers), try_include_seq("Comments", _arrow2896, assay.Comments)]))
    class ObjectExpr2897(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], assay: Any=assay) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr2897()


def _arrow2908(get: IGetters) -> ArcAssay:
    def _arrow2898(__unit: None=None) -> str:
        object_arg: IRequiredGetter = get.Required
        return object_arg.Field("Identifier", string)

    def _arrow2899(__unit: None=None) -> str | None:
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("Title", string)

    def _arrow2900(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("Description", string)

    def _arrow2901(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("MeasurementType", OntologyAnnotation_decoder)

    def _arrow2902(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("TechnologyType", OntologyAnnotation_decoder)

    def _arrow2903(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("TechnologyPlatform", OntologyAnnotation_decoder)

    def _arrow2904(__unit: None=None) -> Array[ArcTable] | None:
        arg_13: Decoder_1[Array[ArcTable]] = resize_array(decoder_1)
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("Tables", arg_13)

    def _arrow2905(__unit: None=None) -> DataMap | None:
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("DataMap", decoder_2)

    def _arrow2906(__unit: None=None) -> Array[Person] | None:
        arg_17: Decoder_1[Array[Person]] = resize_array(decoder_3)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("Performers", arg_17)

    def _arrow2907(__unit: None=None) -> Array[Comment] | None:
        arg_19: Decoder_1[Array[Comment]] = resize_array(decoder_4)
        object_arg_9: IOptionalGetter = get.Optional
        return object_arg_9.Field("Comments", arg_19)

    return ArcAssay.create(_arrow2898(), _arrow2899(), _arrow2900(), _arrow2901(), _arrow2902(), _arrow2903(), _arrow2904(), _arrow2905(), _arrow2906(), _arrow2907())


decoder: Decoder_1[ArcAssay] = object(_arrow2908)

def encoder_compressed(string_table: Any, oa_table: Any, cell_table: Any, assay: ArcAssay) -> IEncodable:
    def chooser(tupled_arg: tuple[str, IEncodable | None], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2912(__unit: None=None, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        value: str = assay.Identifier
        class ObjectExpr2911(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2911()

    def _arrow2914(value_1: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        class ObjectExpr2913(IEncodable):
            def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                return helpers_1.encode_string(value_1)

        return ObjectExpr2913()

    def _arrow2916(value_3: str, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        class ObjectExpr2915(IEncodable):
            def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                return helpers_2.encode_string(value_3)

        return ObjectExpr2915()

    def _arrow2917(oa: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa)

    def _arrow2918(oa_1: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa_1)

    def _arrow2919(oa_2: OntologyAnnotation, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return OntologyAnnotation_encoder(oa_2)

    def _arrow2920(table: ArcTable, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return encoder_compressed_1(string_table, oa_table, cell_table, table)

    def _arrow2921(dm: DataMap, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return encoder_compressed_2(string_table, oa_table, cell_table, dm)

    def _arrow2922(person: Person, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return encoder_6(person)

    def _arrow2923(comment: Comment, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> IEncodable:
        return encoder_7(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("Identifier", _arrow2912()), try_include("Title", _arrow2914, assay.Title), try_include("Description", _arrow2916, assay.Description), try_include("MeasurementType", _arrow2917, assay.MeasurementType), try_include("TechnologyType", _arrow2918, assay.TechnologyType), try_include("TechnologyPlatform", _arrow2919, assay.TechnologyPlatform), try_include_seq("Tables", _arrow2920, assay.Tables), try_include("DataMap", _arrow2921, assay.DataMap), try_include_seq("Performers", _arrow2922, assay.Performers), try_include_seq("Comments", _arrow2923, assay.Comments)]))
    class ObjectExpr2924(IEncodable):
        def Encode(self, helpers_3: IEncoderHelpers_1[Any], string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table, assay: Any=assay) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_3.encode_object(arg)

    return ObjectExpr2924()


def decoder_compressed(string_table: Array[str], oa_table: Array[OntologyAnnotation], cell_table: Array[CompositeCell]) -> Decoder_1[ArcAssay]:
    def _arrow2935(get: IGetters, string_table: Any=string_table, oa_table: Any=oa_table, cell_table: Any=cell_table) -> ArcAssay:
        def _arrow2925(__unit: None=None) -> str:
            object_arg: IRequiredGetter = get.Required
            return object_arg.Field("Identifier", string)

        def _arrow2926(__unit: None=None) -> str | None:
            object_arg_1: IOptionalGetter = get.Optional
            return object_arg_1.Field("Title", string)

        def _arrow2927(__unit: None=None) -> str | None:
            object_arg_2: IOptionalGetter = get.Optional
            return object_arg_2.Field("Description", string)

        def _arrow2928(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_3: IOptionalGetter = get.Optional
            return object_arg_3.Field("MeasurementType", OntologyAnnotation_decoder)

        def _arrow2929(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_4: IOptionalGetter = get.Optional
            return object_arg_4.Field("TechnologyType", OntologyAnnotation_decoder)

        def _arrow2930(__unit: None=None) -> OntologyAnnotation | None:
            object_arg_5: IOptionalGetter = get.Optional
            return object_arg_5.Field("TechnologyPlatform", OntologyAnnotation_decoder)

        def _arrow2931(__unit: None=None) -> Array[ArcTable] | None:
            arg_13: Decoder_1[Array[ArcTable]] = resize_array(decoder_compressed_1(string_table, oa_table, cell_table))
            object_arg_6: IOptionalGetter = get.Optional
            return object_arg_6.Field("Tables", arg_13)

        def _arrow2932(__unit: None=None) -> DataMap | None:
            arg_15: Decoder_1[DataMap] = decoder_compressed_2(string_table, oa_table, cell_table)
            object_arg_7: IOptionalGetter = get.Optional
            return object_arg_7.Field("DataMap", arg_15)

        def _arrow2933(__unit: None=None) -> Array[Person] | None:
            arg_17: Decoder_1[Array[Person]] = resize_array(decoder_3)
            object_arg_8: IOptionalGetter = get.Optional
            return object_arg_8.Field("Performers", arg_17)

        def _arrow2934(__unit: None=None) -> Array[Comment] | None:
            arg_19: Decoder_1[Array[Comment]] = resize_array(decoder_4)
            object_arg_9: IOptionalGetter = get.Optional
            return object_arg_9.Field("Comments", arg_19)

        return ArcAssay.create(_arrow2925(), _arrow2926(), _arrow2927(), _arrow2928(), _arrow2929(), _arrow2930(), _arrow2931(), _arrow2932(), _arrow2933(), _arrow2934())

    return object(_arrow2935)


def ROCrate_genID(a: ArcAssay) -> str:
    match_value: str = a.Identifier
    if match_value == "":
        return "#EmptyAssay"

    else: 
        return ("assays/" + replace(match_value, " ", "_")) + "/"



def ROCrate_encoder(assay_name: str | None, a: ArcAssay) -> IEncodable:
    file_name: str = Assay_fileNameFromIdentifier(a.Identifier)
    processes: FSharpList[Process] = ARCtrl_ArcTables__ArcTables_GetProcesses(a)
    data_files: FSharpList[Data] = get_data(processes)
    def chooser(tupled_arg: tuple[str, IEncodable | None], assay_name: Any=assay_name, a: Any=a) -> tuple[str, IEncodable] | None:
        def mapping(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
            return (tupled_arg[0], v_1)

        return map(mapping, tupled_arg[1])

    def _arrow2939(__unit: None=None, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        value: str = ROCrate_genID(a)
        class ObjectExpr2938(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any]) -> Any:
                return helpers.encode_string(value)

        return ObjectExpr2938()

    class ObjectExpr2940(IEncodable):
        def Encode(self, helpers_1: IEncoderHelpers_1[Any], assay_name: Any=assay_name, a: Any=a) -> Any:
            return helpers_1.encode_string("Assay")

    class ObjectExpr2941(IEncodable):
        def Encode(self, helpers_2: IEncoderHelpers_1[Any], assay_name: Any=assay_name, a: Any=a) -> Any:
            return helpers_2.encode_string("Assay")

    def _arrow2943(__unit: None=None, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        value_3: str = a.Identifier
        class ObjectExpr2942(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any]) -> Any:
                return helpers_3.encode_string(value_3)

        return ObjectExpr2942()

    class ObjectExpr2944(IEncodable):
        def Encode(self, helpers_4: IEncoderHelpers_1[Any], assay_name: Any=assay_name, a: Any=a) -> Any:
            return helpers_4.encode_string(file_name)

    def _arrow2946(value_5: str, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        class ObjectExpr2945(IEncodable):
            def Encode(self, helpers_5: IEncoderHelpers_1[Any]) -> Any:
                return helpers_5.encode_string(value_5)

        return ObjectExpr2945()

    def _arrow2948(value_7: str, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        class ObjectExpr2947(IEncodable):
            def Encode(self, helpers_6: IEncoderHelpers_1[Any]) -> Any:
                return helpers_6.encode_string(value_7)

        return ObjectExpr2947()

    def _arrow2949(oa: OntologyAnnotation, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderPropertyValue(oa)

    def _arrow2950(oa_1: OntologyAnnotation, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa_1)

    def _arrow2951(oa_2: OntologyAnnotation, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return OntologyAnnotation_ROCrate_encoderDefinedTerm(oa_2)

    def _arrow2952(oa_3: Person, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return ROCrate_encoder_1(oa_3)

    def _arrow2953(oa_4: Data, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return ROCrate_encoder_2(oa_4)

    def _arrow2955(__unit: None=None, assay_name: Any=assay_name, a: Any=a) -> Callable[[Process], IEncodable]:
        assay_name_1: str | None = a.Identifier
        def _arrow2954(oa_5: Process) -> IEncodable:
            return ROCrate_encoder_3(assay_name, assay_name_1, oa_5)

        return _arrow2954

    def _arrow2956(comment: Comment, assay_name: Any=assay_name, a: Any=a) -> IEncodable:
        return ROCrate_encoder_4(comment)

    values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("@id", _arrow2939()), ("@type", list_1_1(singleton(ObjectExpr2940()))), ("additionalType", ObjectExpr2941()), ("identifier", _arrow2943()), ("filename", ObjectExpr2944()), try_include("title", _arrow2946, a.Title), try_include("description", _arrow2948, a.Description), try_include("measurementType", _arrow2949, a.MeasurementType), try_include("technologyType", _arrow2950, a.TechnologyType), try_include("technologyPlatform", _arrow2951, a.TechnologyPlatform), try_include_seq("performers", _arrow2952, a.Performers), try_include_list("dataFiles", _arrow2953, data_files), try_include_list("processSequence", _arrow2955(), processes), try_include_seq("comments", _arrow2956, a.Comments), ("@context", context_jsonvalue)]))
    class ObjectExpr2957(IEncodable):
        def Encode(self, helpers_7: IEncoderHelpers_1[Any], assay_name: Any=assay_name, a: Any=a) -> Any:
            def mapping_1(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_7))

            arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_1, values)
            return helpers_7.encode_object(arg)

    return ObjectExpr2957()


def _arrow2967(get: IGetters) -> ArcAssay:
    def _arrow2958(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("identifier", string)

    identifier: str = default_arg(_arrow2958(), create_missing_identifier())
    def mapping(arg_4: FSharpList[Process]) -> Array[ArcTable]:
        a: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D(arg_4)
        return a.Tables

    def _arrow2959(__unit: None=None) -> FSharpList[Process] | None:
        arg_3: Decoder_1[FSharpList[Process]] = list_1_2(ROCrate_decoder_1)
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("processSequence", arg_3)

    tables: Array[ArcTable] | None = map(mapping, _arrow2959())
    def _arrow2960(__unit: None=None) -> str | None:
        object_arg_2: IOptionalGetter = get.Optional
        return object_arg_2.Field("title", string)

    def _arrow2961(__unit: None=None) -> str | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("description", string)

    def _arrow2962(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("measurementType", OntologyAnnotation_ROCrate_decoderPropertyValue)

    def _arrow2963(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("technologyType", OntologyAnnotation_ROCrate_decoderDefinedTerm)

    def _arrow2964(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_6: IOptionalGetter = get.Optional
        return object_arg_6.Field("technologyPlatform", OntologyAnnotation_ROCrate_decoderDefinedTerm)

    def _arrow2965(__unit: None=None) -> Array[Person] | None:
        arg_16: Decoder_1[Array[Person]] = resize_array(ROCrate_decoder_2)
        object_arg_7: IOptionalGetter = get.Optional
        return object_arg_7.Field("performers", arg_16)

    def _arrow2966(__unit: None=None) -> Array[Comment] | None:
        arg_18: Decoder_1[Array[Comment]] = resize_array(ROCrate_decoder_3)
        object_arg_8: IOptionalGetter = get.Optional
        return object_arg_8.Field("comments", arg_18)

    return ArcAssay(identifier, _arrow2960(), _arrow2961(), _arrow2962(), _arrow2963(), _arrow2964(), tables, None, _arrow2965(), _arrow2966())


ROCrate_decoder: Decoder_1[ArcAssay] = object(_arrow2967)

def ISAJson_encoder(study_name: str | None, id_map: Any | None, a: ArcAssay) -> IEncodable:
    def f(a_1: ArcAssay, study_name: Any=study_name, id_map: Any=id_map, a: Any=a) -> IEncodable:
        file_name: str = Assay_fileNameFromIdentifier(a_1.Identifier)
        processes: FSharpList[Process] = ARCtrl_ArcTables__ArcTables_GetProcesses(a_1)
        def encoder_1(oa: OntologyAnnotation, a_1: Any=a_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa)

        encoded_units: tuple[str, IEncodable | None] = try_include_list("unitCategories", encoder_1, get_units(processes))
        def encoder_2(value_1: MaterialAttribute, a_1: Any=a_1) -> IEncodable:
            return encoder_8(id_map, value_1)

        encoded_characteristics: tuple[str, IEncodable | None] = try_include_list("characteristicCategories", encoder_2, get_characteristics(processes))
        def _arrow2968(ps: FSharpList[Process], a_1: Any=a_1) -> IEncodable:
            return encoder_9(id_map, ps)

        encoded_materials: tuple[str, IEncodable | None] = try_include("materials", _arrow2968, Option_fromValueWithDefault(empty(), processes))
        def encoder_3(oa_1: Data, a_1: Any=a_1) -> IEncodable:
            return ISAJson_encoder_1(id_map, oa_1)

        encoced_data_files: tuple[str, IEncodable | None] = try_include_list("dataFiles", encoder_3, get_data(processes))
        units: FSharpList[OntologyAnnotation] = get_units(processes)
        def _arrow2971(__unit: None=None, a_1: Any=a_1) -> IEnumerable_1[Comment]:
            def _arrow2970(__unit: None=None) -> IEnumerable_1[Comment]:
                def _arrow2969(__unit: None=None) -> IEnumerable_1[Comment]:
                    return singleton_1(Comment.create("description", value_9(a_1.Description))) if (a_1.Description is not None) else empty_1()

                return append(singleton_1(Comment.create("title", value_9(a_1.Title))) if (a_1.Title is not None) else empty_1(), delay(_arrow2969))

            return append(a_1.Comments if (len(a_1.Comments) > 0) else empty_1(), delay(_arrow2970))

        comments: FSharpList[Comment] = to_list(delay(_arrow2971))
        def chooser(tupled_arg: tuple[str, IEncodable | None], a_1: Any=a_1) -> tuple[str, IEncodable] | None:
            def mapping_1(v_1: IEncodable, tupled_arg: Any=tupled_arg) -> tuple[str, IEncodable]:
                return (tupled_arg[0], v_1)

            return map(mapping_1, tupled_arg[1])

        class ObjectExpr2973(IEncodable):
            def Encode(self, helpers: IEncoderHelpers_1[Any], a_1: Any=a_1) -> Any:
                return helpers.encode_string(file_name)

        def _arrow2975(value_5: str, a_1: Any=a_1) -> IEncodable:
            class ObjectExpr2974(IEncodable):
                def Encode(self, helpers_1: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_1.encode_string(value_5)

            return ObjectExpr2974()

        def _arrow2976(oa_2: OntologyAnnotation, a_1: Any=a_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa_2)

        def _arrow2977(oa_3: OntologyAnnotation, a_1: Any=a_1) -> IEncodable:
            return OntologyAnnotation_ISAJson_encoder(id_map, oa_3)

        def _arrow2979(value_7: str, a_1: Any=a_1) -> IEncodable:
            class ObjectExpr2978(IEncodable):
                def Encode(self, helpers_2: IEncoderHelpers_1[Any]) -> Any:
                    return helpers_2.encode_string(value_7)

            return ObjectExpr2978()

        def mapping(tp: OntologyAnnotation, a_1: Any=a_1) -> str:
            return JsonTypes_composeTechnologyPlatform(tp)

        def _arrow2981(__unit: None=None, a_1: Any=a_1) -> Callable[[Process], IEncodable]:
            assay_name: str | None = a_1.Identifier
            def _arrow2980(oa_4: Process) -> IEncodable:
                return ISAJson_encoder_2(study_name, assay_name, id_map, oa_4)

            return _arrow2980

        def _arrow2982(comment: Comment, a_1: Any=a_1) -> IEncodable:
            return ISAJson_encoder_3(id_map, comment)

        values: FSharpList[tuple[str, IEncodable]] = choose(chooser, of_array([("filename", ObjectExpr2973()), try_include("@id", _arrow2975, ROCrate_genID(a_1)), try_include("measurementType", _arrow2976, a_1.MeasurementType), try_include("technologyType", _arrow2977, a_1.TechnologyType), try_include("technologyPlatform", _arrow2979, map(mapping, a_1.TechnologyPlatform)), encoced_data_files, encoded_materials, encoded_characteristics, encoded_units, try_include_list("processSequence", _arrow2981(), processes), try_include_seq("comments", _arrow2982, comments)]))
        class ObjectExpr2983(IEncodable):
            def Encode(self, helpers_3: IEncoderHelpers_1[Any], a_1: Any=a_1) -> Any:
                def mapping_2(tupled_arg_1: tuple[str, IEncodable]) -> tuple[str, __A_]:
                    return (tupled_arg_1[0], tupled_arg_1[1].Encode(helpers_3))

                arg: IEnumerable_1[tuple[str, __A_]] = map_1(mapping_2, values)
                return helpers_3.encode_object(arg)

        return ObjectExpr2983()

    if id_map is not None:
        def _arrow2984(a_2: ArcAssay, study_name: Any=study_name, id_map: Any=id_map, a: Any=a) -> str:
            return ROCrate_genID(a_2)

        return encode(_arrow2984, f, a, id_map)

    else: 
        return f(a)



ISAJson_allowedFields: FSharpList[str] = of_array(["@id", "filename", "measurementType", "technologyType", "technologyPlatform", "dataFiles", "materials", "characteristicCategories", "unitCategories", "processSequence", "comments", "@type", "@context"])

def _arrow2990(get: IGetters) -> ArcAssay:
    def _arrow2985(__unit: None=None) -> str | None:
        object_arg: IOptionalGetter = get.Optional
        return object_arg.Field("filename", string)

    identifier: str = default_arg(bind(Assay_tryIdentifierFromFileName, _arrow2985()), create_missing_identifier())
    def mapping(arg_4: FSharpList[Process]) -> Array[ArcTable]:
        a: ArcTables = ARCtrl_ArcTables__ArcTables_fromProcesses_Static_62A3309D(arg_4)
        return a.Tables

    def _arrow2986(__unit: None=None) -> FSharpList[Process] | None:
        arg_3: Decoder_1[FSharpList[Process]] = list_1_2(ISAJson_decoder_1)
        object_arg_1: IOptionalGetter = get.Optional
        return object_arg_1.Field("processSequence", arg_3)

    tables: Array[ArcTable] | None = map(mapping, _arrow2986())
    comments: Array[Comment] | None
    arg_6: Decoder_1[Array[Comment]] = resize_array(ISAJson_decoder_2)
    object_arg_2: IOptionalGetter = get.Optional
    comments = object_arg_2.Field("comments", arg_6)
    def binder(c: Array[Comment]) -> str | None:
        def chooser(x: Comment, c: Any=c) -> str | None:
            if (value_9(x.Name) == "title") if (x.Name is not None) else False:
                return x.Value

            else: 
                return None


        return try_pick(chooser, c)

    title: str | None = bind(binder, comments)
    def binder_1(c_1: Array[Comment]) -> str | None:
        def chooser_1(x_1: Comment, c_1: Any=c_1) -> str | None:
            if (value_9(x_1.Name) == "description") if (x_1.Name is not None) else False:
                return x_1.Value

            else: 
                return None


        return try_pick(chooser_1, c_1)

    description: str | None = bind(binder_1, comments)
    def binder_2(c_2: Array[Comment]) -> Array[Comment] | None:
        def f(x_2: Comment, c_2: Any=c_2) -> bool:
            if x_2.Name is None:
                return True

            elif value_9(x_2.Name) != "title":
                return value_9(x_2.Name) != "description"

            else: 
                return False


        match_value: Array[Comment] = ResizeArray_filter(f, c_2)
        if length(match_value) == 0:
            return None

        else: 
            return match_value


    comments_1: Array[Comment] | None = bind(binder_2, comments)
    def _arrow2987(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_3: IOptionalGetter = get.Optional
        return object_arg_3.Field("measurementType", OntologyAnnotation_ISAJson_decoder)

    def _arrow2988(__unit: None=None) -> OntologyAnnotation | None:
        object_arg_4: IOptionalGetter = get.Optional
        return object_arg_4.Field("technologyType", OntologyAnnotation_ISAJson_decoder)

    def _arrow2989(__unit: None=None) -> OntologyAnnotation | None:
        arg_12: Decoder_1[OntologyAnnotation] = map_2(JsonTypes_decomposeTechnologyPlatform, string)
        object_arg_5: IOptionalGetter = get.Optional
        return object_arg_5.Field("technologyPlatform", arg_12)

    return ArcAssay(identifier, title, description, _arrow2987(), _arrow2988(), _arrow2989(), tables, None, None, comments_1)


ISAJson_decoder: Decoder_1[ArcAssay] = Decode_objectNoAdditionalProperties(ISAJson_allowedFields, _arrow2990)

__all__ = ["encoder", "decoder", "encoder_compressed", "decoder_compressed", "ROCrate_genID", "ROCrate_encoder", "ROCrate_decoder", "ISAJson_encoder", "ISAJson_allowedFields", "ISAJson_decoder"]

